import gzip
import json
import logging
import os
import time
import types
import typing
from collections.abc import Iterator
from concurrent.futures import ThreadPoolExecutor
from typing import Union

import polars as pl
import requests
from cryptography.fernet import Fernet
from openai import OpenAI
from polars import SQLContext
from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
)

from nosible.classes.result_set import ResultSet
from nosible.classes.search import Search
from nosible.classes.search_set import SearchSet
from nosible.classes.snippet_set import SnippetSet
from nosible.classes.web_page import WebPageData
from nosible.utils.json_tools import json_loads
from nosible.utils.question_builder import _get_question
from nosible.utils.rate_limiter import PLAN_RATE_LIMITS, RateLimiter, _rate_limited

# Set up a module‐level logger.
logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.INFO)
logging.basicConfig(level=logging.DEBUG)
logging.disable(logging.CRITICAL)


class Nosible:
    """
    High-level client for the Nosible Search API.

    Parameters
    ----------
    nosible_api_key : str, optional
        API key for the Nosible Search API.
    llm_api_key : str, optional
        API key for LLM-based query expansions.
    openai_base_url : str
        Base URL for the OpenAI-compatible LLM API. (default is OpenRouter's API endpoint)
    sentiment_model : str, optional
        Model to use for sentiment analysis (default is "openai/gpt-4o").
    timeout : int
        Request timeout for HTTP calls.
    retries : int,
        Number of retry attempts for transient HTTP errors.
    concurrency : int,
        Maximum concurrent search requests.
    publish_start : str, optional
        Start date for when the document was published (ISO format).
    publish_end : str, optional
        End date for when the document was published (ISO format).
    visited_start : str, optional
        Start date for when the document was visited by NOSIBLE (ISO format).
    visited_end : str, optional
        End date for when the document was visited by NOSIBLE (ISO format).
    certain : bool, optional
        Only include documents where we are 100% sure of the date.
    include_netlocs : list of str, optional
        List of netlocs (domains) to include in the search. (Max: 50)
    exclude_netlocs : list of str, optional
        List of netlocs (domains) to exclude in the search. (Max: 50)
    include_languages : list of str, optional
        Languages to include in the search. (Max: 50, ISO 639-1 language codes).
    exclude_languages : list of str, optional
        Language codes to exclude in the search (Max: 50, ISO 639-1 language codes).
    include_companies : list of str, optional
        Google KG IDs of public companies to require (Max: 50).
    exclude_companies : list of str, optional
        Google KG IDs of public companies to forbid (Max: 50).
    include_docs : list of str, optional
        URL hashes of docs to include (Max: 50).
    exclude_docs : list of str, optional
        URL hashes of docs to exclude (Max: 50).

    Notes
    -----
    - The `nosible_api_key` is required to access the Nosible Search API.
    - The `llm_api_key` is optional and used for LLM-based query expansions.
    - The `openai_base_url` defaults to OpenRouter's API endpoint.
    - The `sentiment_model` is used for generating query expansions and sentiment analysis.
    - The `timeout`, `retries`, and `concurrency` parameters control the behavior of HTTP requests.

    Examples
    --------
    >>> from nosible import Nosible  # doctest: +SKIP
    >>> nos = Nosible(nosible_api_key="your_api_key_here")  # doctest: +SKIP
    >>> search = nos.search(question="What is Nosible?", n_results=5)  # doctest: +SKIP
    """

    def __init__(
        self,
        nosible_api_key: str = None,
        llm_api_key: str = None,
        openai_base_url: str = "https://openrouter.ai/api/v1",
        sentiment_model: str = "openai/gpt-4o",
        timeout: int = 30,
        retries: int = 5,
        concurrency: int = 10,
        publish_start: str = None,
        publish_end: str = None,
        include_netlocs: list = None,
        exclude_netlocs: list = None,
        visited_start: str = None,
        visited_end: str = None,
        certain: bool = None,
        include_languages: list = None,
        exclude_languages: list = None,
        include_companies: list = None,
        exclude_companies: list = None,
        include_docs: list = None,
        exclude_docs: list = None,
    ) -> None:
        # API Keys
        if nosible_api_key is not None:
            self.nosible_api_key = nosible_api_key
        elif os.getenv("NOSIBLE_API_KEY") is not None:
            try:
                self.nosible_api_key = os.getenv("NOSIBLE_API_KEY")
            except KeyError:
                raise ValueError("Must provide api_key or set $NOSIBLE_API_KEY")
        else:
            # Neither passed in nor in the environment
            raise ValueError("Must provide api_key or set $NOSIBLE_API_KEY")

        self.llm_api_key = llm_api_key or os.getenv("LLM_API_KEY")
        self.openai_base_url = openai_base_url
        self.sentiment_model = sentiment_model
        # Network parameters
        self.timeout = timeout
        self.retries = retries
        self.concurrency = concurrency

        # Initialize Logger
        self.logger = logging.getLogger(__name__)
        logging.getLogger("httpx").setLevel(logging.WARNING)
        logging.getLogger("httpcore").setLevel(logging.WARNING)

        self._limiters = {
            endpoint: [RateLimiter(calls, period) for calls, period in buckets]
            for endpoint, buckets in PLAN_RATE_LIMITS[self._get_user_plan()].items()
        }

        # Define retry decorator
        self._post = retry(
            reraise=True,
            stop=stop_after_attempt(self.retries) | stop_after_delay(self.timeout),
            wait=wait_exponential(multiplier=1, min=1, max=10),
            retry=retry_if_exception_type(requests.exceptions.RequestException),
            before_sleep=before_sleep_log(self.logger, logging.WARNING),
        )(self._post)

        # Wrap _generate_expansions in the same retry logic
        self._generate_expansions = retry(
            reraise=True,
            stop=stop_after_attempt(self.retries) | stop_after_delay(self.timeout),
            wait=wait_exponential(multiplier=1, min=1, max=10),
            retry=retry_if_exception_type(Exception),
            before_sleep=before_sleep_log(self.logger, logging.WARNING),
        )(self._generate_expansions)

        # Thread pool for parallel searches
        self._session = requests.Session()
        self._executor = ThreadPoolExecutor(max_workers=self.concurrency)

        # Headers
        self.headers = {"Accept-Encoding": "gzip", "Content-Type": "application/json", "api-key": self.nosible_api_key}

        # Filters
        self.publish_start = publish_start
        self.publish_end = publish_end
        self.include_netlocs = include_netlocs
        self.exclude_netlocs = exclude_netlocs
        self.include_companies = include_companies
        self.exclude_companies = exclude_companies
        self.visited_start = visited_start
        self.visited_end = visited_end
        self.certain = certain
        self.include_languages = include_languages
        self.exclude_languages = exclude_languages
        self.include_companies = include_companies
        self.exclude_companies = exclude_companies
        self.exclude_docs = exclude_docs
        self.include_docs = include_docs

    def search(
        self,
        *,
        search: Search = None,
        question: str = None,
        expansions: list[str] = None,
        sql_filter: list[str] = None,
        n_results: int = 100,
        n_probes: int = 30,
        n_contextify: int = 128,
        algorithm: str = "hybrid-2",
        autogenerate_expansions: bool = False,
        publish_start: str = None,
        publish_end: str = None,
        include_netlocs: list = None,
        exclude_netlocs: list = None,
        visited_start: str = None,
        visited_end: str = None,
        certain: bool = None,
        include_languages: list = None,
        exclude_languages: list = None,
        include_companies: list = None,
        exclude_companies: list = None,
        include_docs: list = None,
        exclude_docs: list = None,
    ) -> ResultSet:
        """
        Run a single search query.

        If `question` is a string, it is wrapped into a Search with the provided
        parameters; if it is already a Search instance, its fields take precedence.

        Parameters
        ----------
        question : str
            Query string.
        search : Search
            Search object to search with.
        expansions : list of str, optional
            List of LLM‐generated expansions.
        sql_filter : list of str, optional
            SQL‐style filter clauses.
        n_results : int
            Max number of results (max 100).
        n_probes : int
            Number of index shards to probe.
        n_contextify : int
            Context window size per result.
        algorithm : str
            Search algorithm type.
        autogenerate_expansions : bool
            Do you want to generate expansions automatically using a LLM?
        publish_start : str, optional
            Start date for when the document was published (ISO format).
        publish_end : str, optional
            End date for when the document was published (ISO format).
        visited_start : str, optional
            Start date for when the document was visited by NOSIBLE (ISO format).
        visited_end : str, optional
            End date for when the document was visited by NOSIBLE (ISO format).
        certain : bool, optional
            Only include documents where we are 100% sure of the date.
        include_netlocs : list of str, optional
            List of netlocs (domains) to include in the search. (Max: 50)
        exclude_netlocs : list of str, optional
            List of netlocs (domains) to exclude in the search. (Max: 50)
        include_languages : list of str, optional
            Languages to include in the search. (Max: 50, ISO 639-1 language codes).
        exclude_languages : list of str, optional
            Language codes to exclude in the search (Max: 50, ISO 639-1 language codes).
        include_companies : list of str, optional
            Google KG IDs of public companies to require (Max: 50).
        exclude_companies : list of str, optional
            Google KG IDs of public companies to forbid (Max: 50).
        include_docs : list of str, optional
            URL hashes of docs to include (Max: 50).
        exclude_docs : list of str, optional
            URL hashes of docs to exclude (Max: 50).

        Returns
        -------
        ResultSet
            The results of the search.

        Raises
        ------
        TypeError
            If both question and search are specified
        TypeError
            If neither question nor search are specified
        RuntimeError
            If the response fails in any way.
        ValueError
            If `n_results` is greater than 100.

        Notes
        -----
        You must provide either a `question` string or a `Search` object, not both.
        The search parameters will be set from the provided object or string and any additional keyword arguments.
        include_companies and exclude_companies must be the Google KG IDs of public companies.

        Examples
        --------
        >>> from nosible.classes.search import Search
        >>> from nosible import Nosible
        >>> s = Search(question="Hedge funds seek to expand into private credit", n_results=10)
        >>> with Nosible() as nos:
        ...     results = nos.search(search=s)
        ...     print(isinstance(results, ResultSet))
        ...     print(len(results))
        True
        10
        >>> nos = Nosible(nosible_api_key="test|xyz")
        >>> nos.search()  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Specify exactly one of 'question' or 'search'.
        >>> nos = Nosible(nosible_api_key="test|xyz")
        >>> nos.search(question="foo", search=s)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Specify exactly one of 'question' or 'search'.
        >>> nos = Nosible(nosible_api_key="test|xyz")
        >>> nos.search(question="foo", n_results=101)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: Search can not have more than 100 results - Use bulk search instead.
        """
        if (question is None and search is None) or (question is not None and search is not None):
            raise TypeError("Specify exactly one of 'question' or 'search'.")

        search_obj = self._construct_search(
            question=search if search is not None else question,
            expansions=expansions,
            sql_filter=sql_filter,
            n_results=n_results,
            n_probes=n_probes,
            n_contextify=n_contextify,
            algorithm=algorithm,
            autogenerate_expansions=autogenerate_expansions,
            publish_start=publish_start,
            publish_end=publish_end,
            include_netlocs=include_netlocs,
            exclude_netlocs=exclude_netlocs,
            visited_start=visited_start,
            visited_end=visited_end,
            certain=certain,
            include_languages=include_languages,
            exclude_languages=exclude_languages,
            include_companies=include_companies,
            exclude_companies=exclude_companies,
            include_docs=include_docs,
            exclude_docs=exclude_docs,
        )

        future = self._executor.submit(self._search_single, search_obj)
        try:
            return future.result()
        except ValueError:
            # Propagate our own “too many results” error directly
            raise
        except Exception as e:
            self.logger.warning(f"Search for {search_obj.question!r} failed: {e}")
            raise RuntimeError(f"Search for {search_obj.question!r} failed") from e

    def searches(
        self,
        *,
        searches: Union[SearchSet, list[Search]] = None,
        questions: list[str] = None,
        expansions: list[str] = None,
        sql_filter: list[str] = None,
        n_results: int = 100,
        n_probes: int = 30,
        n_contextify: int = 128,
        algorithm: str = "hybrid-2",
        autogenerate_expansions: bool = False,
        publish_start: str = None,
        publish_end: str = None,
        include_netlocs: list = None,
        exclude_netlocs: list = None,
        visited_start: str = None,
        visited_end: str = None,
        certain: bool = None,
        include_languages: list = None,
        exclude_languages: list = None,
        include_companies: list = None,
        exclude_companies: list = None,
        include_docs: list = None,
        exclude_docs: list = None,
    ) -> Iterator[ResultSet]:
        """
        Run multiple searches concurrently and yield results.

        Parameters
        ----------
        searches: SearchSet or list of Search
            The searches execute.
        questions : list of str
            The search queries to execute.
        expansions : list of str, optional
            List of expansion terms to use for each search.
        sql_filter : list of str, optional
            SQL-like filters to apply to the search.
        n_results : int
            Number of results to return per search.
        n_probes : int
            Number of probes to use for the search algorithm.
        n_contextify : int
            Context window size for the search.
        algorithm : str
            Search algorithm to use.
        autogenerate_expansions : bool
            Do you want to generate expansions automatically using a LLM?
        publish_start : str, optional
            Start date for when the document was published (ISO format).
        publish_end : str, optional
            End date for when the document was published (ISO format).
        visited_start : str, optional
            Start date for when the document was visited by NOSIBLE (ISO format).
        visited_end : str, optional
            End date for when the document was visited by NOSIBLE (ISO format).
        certain : bool, optional
            Only include documents where we are 100% sure of the date.
        include_netlocs : list of str, optional
            List of netlocs (domains) to include in the search. (Max: 50)
        exclude_netlocs : list of str, optional
            List of netlocs (domains) to exclude in the search. (Max: 50)
        include_languages : list of str, optional
            Languages to include in the search. (Max: 50, ISO 639-1 language codes).
        exclude_languages : list of str, optional
            Language codes to exclude in the search (Max: 50, ISO 639-1 language codes).
        include_companies : list of str, optional
            Google KG IDs of public companies to require (Max: 50).
        exclude_companies : list of str, optional
            Google KG IDs of public companies to forbid (Max: 50).
        include_docs : list of str, optional
            URL hashes of docs to include (Max: 50).
        exclude_docs : list of str, optional
            URL hashes of docs to exclude (Max: 50).

        Returns
        ------
        ResultSet or None
            Each completed search’s results, or None on failure.

        Raises
        ------
        TypeError
            If `queries` is not a list of strings, a list of Search objects, or a SearchSet instance.
        TypeError
            If both queries and searches are specified.
        TypeError
            If neither queries nor searches are specified.

        Notes
        -----
        You must provide either a list of `questions` or a list of `Search` objects, not both.
        The search parameters will be set from the provided object or string and any additional keyword arguments.

        Examples
        --------
        >>> from nosible import Nosible
        >>> queries = SearchSet(
        ...     [
        ...         Search(question="Hedge funds seek to expand into private credit", n_results=5),
        ...         Search(question="How have the Trump tariffs impacted the US economy?", n_results=5),
        ...     ]
        ... )
        >>> with Nosible() as nos:
        ...     results_list = list(nos.searches(searches=queries))
        >>> print(len(results_list))
        2
        >>> for r in results_list:
        ...     print(isinstance(r, ResultSet), bool(r))
        True True
        True True
        >>> with Nosible() as nos:
        ...     results_list_str = list(
        ...         nos.searches(
        ...             questions=[
        ...                 "What are the terms of the partnership between Microsoft and OpenAI?",
        ...                 "What are the terms of the partnership between Volkswagen and Uber?",
        ...             ]
        ...         )
        ...     )
        >>> nos = Nosible(nosible_api_key="test|xyz")  # doctest: +ELLIPSIS
        >>> nos.searches()  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Specify exactly one of 'questions' or 'searches'.
        >>> from nosible import Nosible
        >>> nos = Nosible(nosible_api_key="test|xyz")
        >>> nos.searches(questions=["A"], searches=SearchSet(searches=["A"]))  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: Specify exactly one of 'questions' or 'searches'.
        """
        if (questions is None and searches is None) or (questions is not None and searches is not None):
            raise TypeError("Specify exactly one of 'questions' or 'searches'.")

        # Function to ensure correct errors are raised.
        def _run_generator():
            search_queries = questions if questions is not None else searches

            searches_list = self._construct_search(
                question=search_queries,
                expansions=expansions,
                sql_filter=sql_filter,
                n_results=n_results,
                n_probes=n_probes,
                n_contextify=n_contextify,
                algorithm=algorithm,
                autogenerate_expansions=autogenerate_expansions,
                publish_start=publish_start,
                publish_end=publish_end,
                include_netlocs=include_netlocs,
                exclude_netlocs=exclude_netlocs,
                visited_start=visited_start,
                visited_end=visited_end,
                certain=certain,
                include_languages=include_languages,
                exclude_languages=exclude_languages,
                include_companies=include_companies,
                exclude_companies=exclude_companies,
                include_docs=include_docs,
                exclude_docs=exclude_docs,
            )

            futures = [self._executor.submit(self._search_single, s) for s in searches_list]

            for future in futures:
                try:
                    yield future.result()
                except Exception as e:
                    self.logger.warning(f"Search failed: {e!r}")
                    yield None

        return _run_generator()

    @_rate_limited("fast")
    def _search_single(self, search_obj: Search) -> ResultSet:
        """
        Execute a single search request using the parameters from a Search object.

        Parameters
        ----------
        search_obj : Search
            A Search instance containing all search parameters.

        Returns
        -------
        ResultSet
            The results of the search.

        Raises
        ------
        ValueError
            If `n_results` > 100.

        Examples
        --------
        >>> from nosible.classes.search import Search
        >>> from nosible import Nosible
        >>> s = Search(question="Nvidia insiders dump more than $1 billion in stock", n_results=200)
        >>> with Nosible() as nos:
        ...     results = nos.search(search=s)  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        ValueError: Search can not have more than 100 results - Use bulk search instead.
        """
        # --------------------------------------------------------------------------------------------------------------
        # Setting search params. Individual search will override Nosible defaults.
        # --------------------------------------------------------------------------------------------------------------
        question = search_obj.question  # No default
        expansions = search_obj.expansions if search_obj.expansions is not None else []  # Default to empty list
        sql_filter = search_obj.sql_filter if search_obj.sql_filter is not None else None
        n_results = search_obj.n_results if search_obj.n_results is not None else 100
        n_probes = search_obj.n_probes if search_obj.n_probes is not None else 30
        n_contextify = search_obj.n_contextify if search_obj.n_contextify is not None else 128
        algorithm = search_obj.algorithm if search_obj.algorithm is not None else "hybrid-2"
        autogenerate_expansions = (
            search_obj.autogenerate_expansions if search_obj.autogenerate_expansions is not None else False
        )
        publish_start = search_obj.publish_start if search_obj.publish_start is not None else self.publish_start
        publish_end = search_obj.publish_end if search_obj.publish_end is not None else self.publish_end
        include_netlocs = search_obj.include_netlocs if search_obj.include_netlocs is not None else self.include_netlocs
        exclude_netlocs = search_obj.exclude_netlocs if search_obj.exclude_netlocs is not None else self.exclude_netlocs
        visited_start = search_obj.visited_start if search_obj.visited_start is not None else self.visited_start
        visited_end = search_obj.visited_end if search_obj.visited_end is not None else self.visited_end
        certain = search_obj.certain if search_obj.certain is not None else self.certain
        include_languages = (
            search_obj.include_languages if search_obj.include_languages is not None else self.include_languages
        )
        exclude_languages = (
            search_obj.exclude_languages if search_obj.exclude_languages is not None else self.exclude_languages
        )
        include_companies = (
            search_obj.include_companies if search_obj.include_companies is not None else self.include_companies
        )
        exclude_companies = (
            search_obj.exclude_companies if search_obj.exclude_companies is not None else self.exclude_companies
        )

        # Generate expansions if not provided
        if expansions is None:
            expansions = []
        if autogenerate_expansions is True:
            expansions = self._generate_expansions(question=question)

        # Generate sql_filter if not provided
        if sql_filter is None:
            sql_filter = self._format_sql(
                publish_start=publish_start,
                publish_end=publish_end,
                include_netlocs=include_netlocs,
                exclude_netlocs=exclude_netlocs,
                visited_start=visited_start,
                visited_end=visited_end,
                certain=certain,
                include_languages=include_languages,
                exclude_languages=exclude_languages,
                include_companies=include_companies,
                exclude_companies=exclude_companies,
            )

        # Enforce limits
        if n_results > 100:
            raise ValueError("Search can not have more than 100 results - Use bulk search instead.")
        filter_responses = n_results
        n_results = max(n_results, 10)

        payload = {
            "question": question,
            "expansions": expansions,
            "sql_filter": sql_filter,
            "n_results": n_results,
            "n_probes": n_probes,
            "n_contextify": n_contextify,
            "algorithm": algorithm,
        }

        resp = self._post(url="https://www.nosible.ai/search/v1/fast-search", payload=payload)
        resp.raise_for_status()
        items = resp.json().get("response", [])[:filter_responses]
        return ResultSet.from_dicts(items)

    @staticmethod
    def _construct_search(
        question: Union[str, Search, SearchSet, list[Search], list[str]], **options
    ) -> Union[Search, SearchSet]:
        """
        Constructs a `Search` or `SearchSet` object from the provided input.
        Parameters
        ----------
        question : Union[str, Search, SearchSet, list[Search], list[str]]
            The input to construct the search from. This can be a single search query string,
            a `Search` object, a `SearchSet` object, or a list of either search query strings or `Search` objects.
        **options
            Additional keyword arguments to pass to the `Search` initializer.
        Returns
        -------
        Union[Search, SearchSet]
            A `Search` object if the input is a single query or `Search`, or a `SearchSet` object if the input is a
            list or a `SearchSet`.
        Raises
        ------
        TypeError
            If `question` is not a `str`, `Search`, `SearchSet`, or a list of these types.
        Notes
        -----
        All extra parameters are passed through to the `Search` initializer.
        """

        def make_search(q: Union[str, Search]) -> Search:
            return q if isinstance(q, Search) else Search(question=q, **options)

        if isinstance(question, SearchSet):
            return question
        if isinstance(question, Search):
            return question
        if isinstance(question, list):
            return SearchSet([make_search(q) for q in question])
        if isinstance(question, str):
            return make_search(question)

        raise TypeError("`question` must be str, Search, SearchSet, or a list thereof")

    @_rate_limited("slow")
    def bulk_search(
        self,
        *,
        search: Search = None,
        question: str = None,
        expansions: list[str] = None,
        sql_filter: list[str] = None,
        n_results: int = 1000,
        n_probes: int = 30,
        n_contextify: int = 128,
        algorithm: str = "hybrid-2",
        autogenerate_expansions: bool = False,
        publish_start: str = None,
        publish_end: str = None,
        include_netlocs: list = None,
        exclude_netlocs: list = None,
        visited_start: str = None,
        visited_end: str = None,
        certain: bool = None,
        include_languages: list = None,
        exclude_languages: list = None,
        include_companies: list = None,
        exclude_companies: list = None,
        include_docs: list = None,
        exclude_docs: list = None,
        verbose: bool = False,
    ) -> ResultSet:
        """
        Perform a bulk (slow) search query (1,000–10,000 results) against the Nosible API.

        Parameters
        ----------
        question : str or None
            Query string. Provide either a question string or a Search object.
        search : Search or None
            Search object to search with. Provide either a Search object or a question string.
        expansions : list of str, optional
            Optional list of expanded query strings.
        sql_filter : list of str, optional
            Optional SQL WHERE clause filters.
        n_results : int
            Number of results per query (1,000–10,000).
        n_probes : int
            Number of shards to probe.
        n_contextify : int
            Context window size per result.
        algorithm : str
            Search algorithm identifier.
        autogenerate_expansions : bool
            Do you want to generate expansions automatically using a LLM?
        publish_start : str, optional
            Start date for when the document was published (ISO format).
        publish_end : str, optional
            End date for when the document was published (ISO format).
        visited_start : str, optional
            Start date for when the document was visited by NOSIBLE (ISO format).
        visited_end : str, optional
            End date for when the document was visited by NOSIBLE (ISO format).
        certain : bool, optional
            Only include documents where we are 100% sure of the date.
        include_netlocs : list of str, optional
            List of netlocs (domains) to include in the search. (Max: 50)
        exclude_netlocs : list of str, optional
            List of netlocs (domains) to exclude in the search. (Max: 50)
        include_languages : list of str, optional
            Languages to include in the search. (Max: 50, ISO 639-1 language codes).
        exclude_languages : list of str, optional
            Language codes to exclude in the search (Max: 50, ISO 639-1 language codes).
        include_companies : list of str, optional
            Google KG IDs of public companies to require (Max: 50).
        exclude_companies : list of str, optional
            Google KG IDs of public companies to forbid (Max: 50).
        include_docs : list of str, optional
            URL hashes of docs to include (Max: 50).
        exclude_docs : list of str, optional
            URL hashes of docs to exclude (Max: 50).
        verbose : bool, optional
            Show verbose output, Bulk search will print more information.

        Returns
        -------
        ResultSet
            The results of the bulk search.

        Raises
        ------
        ValueError
            If `n_results` is out of bounds (<1000 or >10000).
        TypeError
            If both question and search are specified.
        TypeError
            If neither question nor search are specified.
        RuntimeError
            If the response fails in any way.

        Notes
        -----
        You must provide either a `question` string or a `Search` object, not both.
        The search parameters will be set from the provided object or string and any additional keyword arguments.

        Examples
        --------
        >>> from nosible.classes.search import Search  # doctest: +SKIP
        >>> from nosible import Nosible  # doctest: +SKIP
        >>> with Nosible(exclude_netlocs=["bbc.com"]) as nos:  # doctest: +SKIP
        ...     results = nos.bulk_search(question=_get_question(), n_results=2000)  # doctest: +SKIP
        ...     print(isinstance(results, ResultSet))  # doctest: +SKIP
        ...     print(len(results))  # doctest: +SKIP
        True
        2000
        >>> s = Search(question=_get_question(), n_results=1000)  # doctest: +SKIP
        >>> with Nosible() as nos:  # doctest: +SKIP
        ...     results = nos.bulk_search(search=s)  # doctest: +SKIP
        ...     print(isinstance(results, ResultSet))  # doctest: +SKIP
        ...     print(len(results))  # doctest: +SKIP
        True
        1000
        >>> nos = Nosible(nosible_api_key="test|xyz")  # doctest: +SKIP
        >>> nos.bulk_search()  # doctest: +SKIP
        Traceback (most recent call last):
        ...
        TypeError: Either question or search must be specified

        >>> nos = Nosible(nosible_api_key="test|xyz")  # doctest: +SKIP
        >>> nos.bulk_search(question=_get_question(), search=Search(question=_get_question()))  # doctest: +SKIP
        Traceback (most recent call last):
        ...
        TypeError: Question and search cannot be both specified
        >>> nos = Nosible(nosible_api_key="test|xyz")  # doctest: +SKIP
        >>> nos.bulk_search(question=_get_question(), n_results=100)  # doctest: +SKIP
        Traceback (most recent call last):
        ...
        ValueError: Bulk search must have at least 1000 results per query; use search() for smaller result sets.
        >>> nos = Nosible(nosible_api_key="test|xyz")  # doctest: +SKIP
        >>> nos.bulk_search(question=_get_question(), n_results=10001)  # doctest: +SKIP
        Traceback (most recent call last):  # doctest: +SKIP
        ...
        ValueError: Bulk search cannot have more than 10000 results per query.
        """
        previous_level = self.logger.level
        if verbose:
            self.logger.setLevel(logging.INFO)

        if question is not None and search is not None:
            raise TypeError("Question and search cannot be both specified")

        if question is None and search is None:
            raise TypeError("Either question or search must be specified")

        # If a Search object is provided, extract its fields
        if search is not None:
            question = search.question
            expansions = search.expansions if search.expansions is not None else expansions
            sql_filter = search.sql_filter if search.sql_filter is not None else sql_filter
            n_results = search.n_results if search.n_results is not None else n_results
            n_probes = search.n_probes if search.n_probes is not None else n_probes
            n_contextify = search.n_contextify if search.n_contextify is not None else n_contextify
            algorithm = search.algorithm if search.algorithm is not None else algorithm
            autogenerate_expansions = (
                search.autogenerate_expansions
                if search.autogenerate_expansions is not None
                else autogenerate_expansions
            )
            publish_start = search.publish_start if search.publish_start is not None else publish_start
            publish_end = search.publish_end if search.publish_end is not None else publish_end
            include_netlocs = search.include_netlocs if search.include_netlocs is not None else include_netlocs
            exclude_netlocs = search.exclude_netlocs if search.exclude_netlocs is not None else exclude_netlocs
            visited_start = search.visited_start if search.visited_start is not None else visited_start
            visited_end = search.visited_end if search.visited_end is not None else visited_end
            certain = search.certain if search.certain is not None else certain
            include_languages = search.include_languages if search.include_languages is not None else include_languages
            exclude_languages = search.exclude_languages if search.exclude_languages is not None else exclude_languages
            include_companies = search.include_companies if search.include_companies is not None else include_companies
            exclude_companies = search.exclude_companies if search.exclude_companies is not None else exclude_companies
            include_docs = search.include_docs if search.include_docs is not None else include_docs
            exclude_docs = search.exclude_docs if search.exclude_docs is not None else exclude_docs

        # Default expansions and filters
        if expansions is None:
            expansions = []
        if autogenerate_expansions is True:
            expansions = self._generate_expansions(question=question)

        # Generate sql_filter if unset
        if sql_filter is None:
            sql_filter = self._format_sql(
                publish_start=publish_start if publish_start is not None else self.publish_start,
                publish_end=publish_end if publish_end is not None else self.publish_end,
                include_netlocs=include_netlocs if include_netlocs is not None else self.include_netlocs,
                exclude_netlocs=exclude_netlocs if exclude_netlocs is not None else self.exclude_netlocs,
                visited_start=visited_start if visited_start is not None else self.visited_start,
                visited_end=visited_end if visited_end is not None else self.visited_end,
                certain=certain if certain is not None else self.certain,
                include_languages=include_languages if include_languages is not None else self.include_languages,
                exclude_languages=exclude_languages if exclude_languages is not None else self.exclude_languages,
                include_companies=include_companies if include_companies is not None else self.include_companies,
                exclude_companies=exclude_companies if exclude_companies is not None else self.exclude_companies,
                include_docs=include_docs if include_docs is not None else self.include_docs,
                exclude_docs=exclude_docs if exclude_docs is not None else self.exclude_docs,
            )

        self.logger.debug(f"SQL Filter: {sql_filter}")

        # Validate n_result bounds
        if n_results < 1000:
            raise ValueError(
                "Bulk search must have at least 1000 results per query; use search() for smaller result sets."
            )
        if n_results > 10000:
            raise ValueError("Bulk search cannot have more than 10000 results per query.")

        # Enforce Minimums
        filter_responses = n_results
        # Slow search must ask for at least 1 000
        n_results = max(n_results, 1000)

        self.logger.info(f"Performing bulk search for {question!r}...")

        try:
            payload = {
                "question": question,
                "expansions": expansions,
                "sql_filter": sql_filter,
                "n_results": n_results,
                "n_probes": n_probes,
                "n_contextify": n_contextify,
                "algorithm": algorithm,
            }
            resp = self._post(url="https://www.nosible.ai/search/v1/slow-search", payload=payload)
            try:
                resp.raise_for_status()
            except requests.HTTPError as e:
                raise ValueError(f"[{question!r}] HTTP {resp.status_code}: {resp.text}") from e

            data = resp.json()

            # Slow search: download & decrypt
            download_from = data.get("download_from")
            if ".zstd." in download_from:
                download_from = download_from.replace(".zstd.", ".gzip.", 1)
            decrypt_using = data.get("decrypt_using")
            for _ in range(100):
                dl = self._session.get(download_from, timeout=self.timeout)
                if dl.ok:
                    fernet = Fernet(decrypt_using.encode())
                    decrypted = fernet.decrypt(dl.content)
                    decompressed = gzip.decompress(decrypted)
                    api_resp = json_loads(decompressed)
                    return ResultSet.from_dicts(api_resp.get("response", [])[:filter_responses])
                time.sleep(10)
            raise ValueError("Results were not retrieved from Nosible")
        except Exception as e:
            self.logger.warning(f"Bulk search for {question!r} failed: {e}")
            raise RuntimeError(f"Bulk search for {question!r} failed") from e
        finally:
            # Restore whatever logging level we had before
            if verbose:
                self.logger.setLevel(previous_level)

    @_rate_limited("visit")
    def visit(self, html: str = "", recrawl: bool = False, render: bool = False, url: str = None) -> WebPageData:
        """
        Visit a given URL and return a structured WebPageData object for the page.

        Parameters
        ----------
        html : str
            Raw HTML to process instead of fetching.
        recrawl : bool
            If True, force a fresh crawl.
        render : bool
            If True, allow JavaScript rendering before extraction.
        url : str
            The URL to fetch and parse.

        Returns
        -------
        WebPageData
            Structured page data object.

        Raises
        ------
        TypeError
            If URL is not provided.
        ValueError
            If invalid JSON response from the server.
        ValueError
            If URL is not found.
        ValueError
            If the server did not send back a 'response' key.

        Examples
        --------
        >>> from nosible import Nosible
        >>> with Nosible() as nos:
        ...     out = nos.visit(url="https://www.dailynewsegypt.com/2023/09/08/g20-and-its-summits/")
        ...     print(isinstance(out, WebPageData))
        ...     print(hasattr(out, "languages"))
        ...     print(hasattr(out, "page"))
        True
        True
        True
        >>> with Nosible() as nos:
        ...     out = nos.visit()
        ...     print(isinstance(out, type(WebPageData)))
        ...     print(hasattr(out, "languages"))
        ...     print(hasattr(out, "page"))  # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ...
        TypeError: URL must be provided
        """
        if url is None:
            raise TypeError("URL must be provided")
        response = self._post(
            url="https://www.nosible.ai/search/v1/visit",
            payload={"html": html, "recrawl": recrawl, "render": render, "url": url},
        )
        try:
            data = response.json()
        except Exception as e:
            self.logger.error(f"Failed to parse JSON from response: {e}")
            raise ValueError("Invalid JSON response from server") from e

        if data == {"message": "Sorry, the URL could not be fetched."}:
            raise ValueError("The URL could not be found.")

        if "response" not in data:
            self.logger.error(f"No 'response' key in server response: {data}")
            raise ValueError("No 'response' key in server response")

        response_data = data["response"]
        return WebPageData(
            companies=response_data.get("companies"),
            full_text=response_data.get("full_text"),
            languages=response_data.get("languages"),
            metadata=response_data.get("metadata"),
            page=response_data.get("page"),
            request=response_data.get("request"),
            snippets=SnippetSet.from_dict(response_data.get("snippets", {})),
            statistics=response_data.get("statistics"),
            structured=response_data.get("structured"),
            url_tree=response_data.get("url_tree"),
        )

    def version(self) -> str:
        """
        Retrieve the current version information for the Nosible API.

        Returns
        -------
        str
            JSON-formatted string containing API version details.

        Examples
        --------
        >>> import json
        >>> from nosible import Nosible
        >>> with Nosible() as nos:
        ...     v = nos.version()
        ...     data = json.loads(v)
        ...     # top‐level object must be a dict
        ...     print(isinstance(data, dict))
        ...     # must have a "response" key mapping to another dict
        ...     print("response" in data and isinstance(data["response"], dict))
        ...     # that inner dict must have exactly the expected sub-keys
        ...     expected = {
        ...         "database",
        ...         "date",
        ...         "documents",
        ...         "runtime",
        ...         "snippets",
        ...         "time",
        ...         "tokens",
        ...         "version",
        ...         "words",
        ...     }
        ...     print(set(data["response"].keys()) == expected)
        True
        True
        True
        """
        response = self._post(url="https://www.nosible.ai/search/v1/version", payload={})

        return json.dumps(response.json(), indent=2, sort_keys=True)

    def indexed(self, url: str = None) -> bool:
        """
        This function checks if a URL has been indexed by Nosible.

        Parameters
        ----------
        url : str, optional
            The full URL to verify.

        Returns
        -------
        bool
            True if the URL is in the index.
            False if the URL is not in the index.

        Raises
        ------

        Examples
        --------
        >>> from nosible import Nosible
        >>> with Nosible() as nos:
        ...     print(nos.indexed(url="https://www.dailynewsegypt.com/2023/09/08/g20-and-its-summits/"))
        True
        """
        response = self._post(url="https://www.nosible.ai/search/v1/indexed", payload={"url": url})

        try:
            response.raise_for_status()
            data = response.json()
            msg = data.get("message")
            if msg == "The URL is in the system.":
                return True
            if msg == "The URL is nowhere to be found.":
                return False
            if msg == "The URL could not be retrieved.":
                return False
        except requests.HTTPError:
            return False
        except:
            return False

    def preflight(self, url: str = None) -> str:
        """
        Run a preflight check for crawling/preprocessing on a URL.

        Parameters
        ----------
        url : str, optional
            The URL to validate or prepare for indexing.

        Returns
        -------
        str
            JSON-formatted string with errors, warnings, or recommendations.

        Examples
        --------
        >>> from nosible import Nosible
        >>> with Nosible() as nos:
        ...     pf = nos.preflight(url="https://www.dailynewsegypt.com/2023/09/08/g20-and-its-summits/")
        ...     print(pf)
        {
          "response": {
            "domain": "dailynewsegypt",
            "fragment": "",
            "geo": "US",
            "hash": "ENNmqkF1mGNhVhvhmbUEs4U2",
            "netloc": "www.dailynewsegypt.com",
            "path": "/2023/09/08/g20-and-its-summits/",
            "prefix": "www",
            "proxy": "US",
            "query": "",
            "query_allowed": {},
            "query_blocked": {},
            "raw_url": "https://www.dailynewsegypt.com/2023/09/08/g20-and-its-summits/",
            "scheme": "https",
            "suffix": "com",
            "url": "https://www.dailynewsegypt.com/2023/09/08/g20-and-its-summits"
          }
        }
        """
        response = self._post(url="https://www.nosible.ai/search/v1/preflight", payload={"url": url})

        return json.dumps(response.json(), indent=2, sort_keys=True)

    def get_rate_limits(self) -> str:
        """
        Generate a plaintext summary of rate limits for every subscription plan.

        Returns
        -------
        str
            A multi-line string containing rate limits for each plan.

        Examples
        --------
        >>> nos = Nosible(nosible_api_key="test|xyz")
        >>> print(nos.get_rate_limits())  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Below are the rate limits for all NOSIBLE plans.
        To upgrade your package, visit https://www.nosible.ai/products.
        <BLANKLINE>
        Unless otherwise indicated, bulk searches are limited to one-at-a-time per API key.
        <BLANKLINE>
        Free: (Your current plan)
        | Endpoint    | Per Month | Per Minute | Effective CPM |
        | ----------- | --------- | ---------- | ------------- |
        | Search      |      3000 |         60 |         $4.00 |
        | URL Visits  |       300 |         60 |         $4.00 |
        | Bulk Search |       300 |         60 |         $4.00 |
        <BLANKLINE>
        Basic ($49p/m):
        | Endpoint    | Per Month | Per Minute | Effective CPM |
        ...
        """
        # Human-friendly plan names
        display = {
            "test": "Free",
            "basic": "Basic ($49p/m)",
            "pro": "Pro ($199p/m)",
            "pro+": "Pro+ ($799p/m)",
            "bus": "Business ($3999p/m)",
            "bus+": "Business+ ($7499p/m)",
            "ent": "Enterprise ($14999p/m)",
        }

        # Human-friendly endpoint names
        endpoint_name = {"fast": "Search", "visit": "URL Visits", "slow": "Bulk Search"}

        out = [
            "Below are the rate limits for all NOSIBLE plans.",
            "To upgrade your package, visit https://www.nosible.ai/products.\n",
            "Unless otherwise indicated, bulk searches are limited to one-at-a-time per API key.\n"
        ]

        user_plan = self._get_user_plan()
        current_plan = ""
        cpm_counter = 4.0

        # Preserve the order you care about:
        for plan in ["test", "basic", "pro", "pro+", "bus", "bus+", "ent"]:
            name = display.get(plan, plan)
            if plan == user_plan:
                current_plan = " (Your current plan)"

            out.append(f"{name}:{current_plan}")
            out.append("| Endpoint    | Per Month | Per Minute | Effective CPM |")
            out.append("| ----------- | --------- | ---------- | ------------- |")

            for ep in ["fast", "visit", "slow"]:
                buckets = PLAN_RATE_LIMITS[plan][ep]
                # Find minute & day
                minute = next(limit for limit, i in buckets if i == 60)
                month = next(limit for limit, i in buckets if i == 24 * 3600 * 30)
                cpm = f"${cpm_counter:.2f}"

                out.append(f"| {endpoint_name[ep]:<11} | {month:>9} | {minute:>10} | {cpm:>13} |")

            cpm_counter = cpm_counter - 0.5
            out.append("")  # Blank line
            current_plan = ""

        return "\n".join(out)

    def close(self):
        """
        Close the Nosible client, shutting down the HTTP session
        and thread pool to release network and threading resources.

        Examples
        --------
        >>> from nosible import Nosible
        >>> nos = Nosible()
        >>> result = nos.close()
        >>> print(result is None)
        True
        >>> # Calling close again should be a no-op
        >>> nos.close()
        >>> print("No Error")
        No Error
        """
        # Shut down HTTP session
        try:
            self._session.close()
        except Exception:
            pass
        # Shut down thread pool
        try:
            # wait = True ensures all submitted tasks complete or are cancelled
            self._executor.shutdown(wait=True)
        except Exception:
            pass

    def _post(self, url: str, payload: dict, headers: dict = None, timeout: int = None) -> requests.Response:
        """
        Internal helper to send a POST request with retry logic.

        Parameters
        ----------
        url : str
            Endpoint URL.
        payload : dict
            JSON-serializable payload.
        headers : dict, optional
            Override headers for this request.
        timeout : int, optional
            Override timeout for this request.

        Raises
        ------
        ValueError
            If the user API key is invalid.
        ValueError
            If the user hits their rate limit.
        ValueError
            If the user is making too many concurrent searches.
        ValueError
            If an unexpected error occurs.
        ValueError
            If NOSIBLE is currently restarting.
        ValueError
            If NOSIBLE is currently overloaded.

        Returns
        -------
        requests.Response
            The HTTP response object.
        """
        response = self._session.post(
            url=url,
            json=payload,
            headers=headers if headers is not None else self.headers,
            timeout=timeout if timeout is not None else self.timeout,
        )

        # If unauthorized, or if the payload is string too short, treat as invalid API key
        if response.status_code == 401:
            raise ValueError("Your API key is not valid.")
        if response.status_code == 422:
            # Only inspect JSON if it’s a JSON response
            content_type = response.headers.get("Content-Type", "")
            if content_type.startswith("application/json"):
                body = response.json()
                if body.get("type") == "string_too_short":
                    raise ValueError("Your API key is not valid: Too Short.")
            else:
                raise ValueError("You made a bad request.")
        if response.status_code == 429:
            raise ValueError("You have hit your rate limit.")
        if response.status_code == 409:
            raise ValueError("Too many concurrent searches.")
        if response.status_code == 500:
            raise ValueError("An unexpected error occurred.")
        if response.status_code == 502:
            raise ValueError("NOSIBLE is currently restarting.")
        if response.status_code == 504:
            raise ValueError("NOSIBLE is currently overloaded.")

        return response

    def _get_user_plan(self) -> str:
        """
        Determine the user's subscription plan from the API key.

        The `nosible_api_key` is expected to start with a plan prefix followed by
        a pipe (`|`) and any additional data. This method splits on the first
        pipe character, validates the prefix against supported plans, and returns it.

        Returns
        -------
        str
            The plan you are currently on.

        Raises
        ------
        ValueError
            If the extracted prefix is not one of the recognized plan names.

        Examples
        --------
        >>> nos = Nosible(nosible_api_key="test+|xyz")  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        ValueError: Your API key is not valid: test+ is not a valid plan prefix.
        """
        # Split off anything after the first '|'
        prefix = (self.nosible_api_key or "").split("|", 1)[0]

        # Map prefixes -> plan names
        plans = {"test", "basic", "pro", "pro+", "bus", "bus+", "ent", "chat"}

        if prefix not in plans:
            raise ValueError(f"Your API key is not valid: {prefix} is not a valid plan prefix.")

        return prefix

    def _generate_expansions(self, question: Union[str, Search]) -> list:
        """
        Generate up to 10 semantically diverse question expansions using an LLM.

        Parameters
        ----------
        question : str
            Original user query.

        Returns
        -------
        list of str
            Up to 10 expanded query strings.

        Raises
        ------
        ValueError
            If no LLM API key is set.
        RuntimeError
            If the LLM response is invalid or cannot be parsed.

        Examples
        --------
        >>> from nosible import Nosible
        >>> nos = Nosible(llm_api_key=None)
        >>> nos.llm_api_key = None
        >>> nos._generate_expansions("anything")  # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        Traceback (most recent call last):
        ...
        ValueError: LLM API key is required for generating expansions.
        """
        if not self.llm_api_key:
            raise ValueError("LLM API key is required for generating expansions.")

        # If the user put in a search, get the question out of it.
        if isinstance(question, Search):
            question = question.question

        # Build a clear prompt that demands JSON output of exactly 10 strings.
        prompt = f"""
            # TASK DESCRIPTION

            Given a search question you must generate a list of 10 similar questions that have the same exact
            semantic meaning but are contextually and lexically different to improve search recall.

            ## Question

            Here is the question you must generate expansions for:

            Question: {question}

            # RESPONSE FORMAT

            Your response must be a JSON object structured as follows: a list of ten strings. Each string must
            be a grammatically correct question that expands on the original question to improve recall.

            [
                string,
                string,
                string,
                string,
                string,
                string,
                string,
                string,
                string,
                string
            ]

            # EXPANSION GUIDELINES

            1. **Use specific named entities** - To improve the quality of your search results you must mention
               specific named entities (people, locations, organizations, products, places) in expansions.

            2. **Expansions must be highly targeted** - To improve the quality of search results each expansion
               must be semantically unambiguous. Questions must be use between ten and fifteen words.

            3. **Expansions must improve recall** - When expanding the question leverage semantic and contextual
               expansion to maximize the ability of the search engine to find semantically relevant documents:

               - Semantic Example: Swap "climate change" with "global warming" or "environmental change".
               - Contextual Example: Swap "diabetes treatment" with "insulin therapy" or "blood sugar management".

        """.replace("                ", "")

        client = OpenAI(base_url=self.openai_base_url, api_key=self.llm_api_key)

        # Call the chat completions endpoint.
        resp = client.chat.completions.create(
            model=self.sentiment_model, messages=[{"role": "user", "content": prompt.strip()}], temperature=0.7
        )

        raw = resp.choices[0].message.content

        # Strip any leading/trailing markdown ``` or text.
        if raw.startswith("```"):
            # remove ```json ... ```
            raw = raw.strip("`").strip()
            # remove optional leading "json"
            if raw.lower().startswith("json"):
                raw = raw[len("json") :].strip()

        # Parse JSON.
        try:
            expansions = json.loads(raw)
        except Exception as decode_err:
            raise RuntimeError(f"OpenRouter response was not valid JSON: '{raw}'") from decode_err

        # Validate.
        if not isinstance(expansions, list) or len(expansions) != 10 or not all(isinstance(q, str) for q in expansions):
            raise RuntimeError("Invalid response: 'choices' missing or empty")

        self.logger.debug(f"Successful expansions: {expansions}")
        return expansions

    def _format_sql(
        self,
        publish_start: str = None,
        publish_end: str = None,
        include_netlocs: list = None,
        exclude_netlocs: list = None,
        visited_start: str = None,
        visited_end: str = None,
        certain: bool = None,
        include_languages: list = None,
        exclude_languages: list = None,
        include_companies: list = None,
        exclude_companies: list = None,
        include_docs: list = None,
        exclude_docs: list = None,
    ) -> str:
        """
        Construct an SQL SELECT statement with WHERE clauses based on provided filters.

        Parameters
        ----------
        publish_start : str, optional
            Start date for when the document was published (ISO format).
        publish_end : str, optional
            End date for when the document was published (ISO format).
        visited_start : str, optional
            Start date for when the document was visited by NOSIBLE (ISO format).
        visited_end : str, optional
            End date for when the document was visited by NOSIBLE (ISO format).
        certain : bool, optional
            Only include documents where we are 100% sure of the date.
        include_netlocs : list of str, optional
            List of netlocs (domains) to include in the search. (Max: 50)
        exclude_netlocs : list of str, optional
            List of netlocs (domains) to exclude in the search. (Max: 50)
        include_languages : list of str, optional
            Languages to include in the search. (Max: 50, ISO 639-1 language codes).
        exclude_languages : list of str, optional
            Language codes to exclude in the search (Max: 50, ISO 639-1 language codes).
        include_companies : list of str, optional
            Google KG IDs of public companies to require (Max: 50).
        exclude_companies : list of str, optional
            Google KG IDs of public companies to forbid (Max: 50).
        include_docs : list of str, optional
            URL hashes of docs to include (Max: 50).
        exclude_docs : list of str, optional
            URL hashes of docs to exclude (Max: 50).

        Returns
        -------
        str
            An SQL query string with appropriate WHERE clauses.

        Raises
        ------
        ValueError
            If more than 50 items in a filter are given.
        """
        # Validate list lengths
        for name, lst in [
            ("include_netlocs", include_netlocs),
            ("exclude_netlocs", exclude_netlocs),
            ("include_languages", include_languages),
            ("exclude_languages", exclude_languages),
            ("include_companies", include_companies),
            ("exclude_companies", exclude_companies),
            ("include_docs", include_docs),
            ("exclude_docs", exclude_docs),
        ]:
            if lst is not None and len(lst) > 50:
                raise ValueError(f"Too many items for '{name}' filter ({len(lst)}); maximum allowed is 50.")

        sql = ["SELECT loc FROM engine"]
        clauses: list[str] = []

        # Published date range
        if publish_start or publish_end:
            if publish_start and publish_end:
                clauses.append(f"published >= '{publish_start}' AND published <= '{publish_end}'")
            elif publish_start:
                clauses.append(f"published >= '{publish_start}'")
            else:  # Only publish_end
                clauses.append(f"published <= '{publish_end}'")

        # Visited date range
        if visited_start or visited_end:
            if visited_start and visited_end:
                clauses.append(f"visited >= '{visited_start}' AND visited <= '{visited_end}'")
            elif visited_start:
                clauses.append(f"visited >= '{visited_start}'")
            else:  # Only visited_end
                clauses.append(f"visited <= '{visited_end}'")

        # date certainty filter
        if certain is True:
            clauses.append("certain = TRUE")
        elif certain is False:
            clauses.append("certain = FALSE")

        # Include netlocs with both www/non-www variants
        if include_netlocs:
            variants = set()
            for n in include_netlocs:
                variants.add(n)
                if n.startswith("www."):
                    variants.add(n[4:])
                else:
                    variants.add("www." + n)
            in_list = ", ".join(f"'{v}'" for v in sorted(variants))
            clauses.append(f"netloc IN ({in_list})")

        # Exclude netlocs with both www/non-www variants
        if exclude_netlocs:
            variants = set()
            for n in exclude_netlocs:
                variants.add(n)
                if n.startswith("www."):
                    variants.add(n[4:])
                else:
                    variants.add("www." + n)
            ex_list = ", ".join(f"'{v}'" for v in sorted(variants))
            clauses.append(f"netloc NOT IN ({ex_list})")

        # Include / exclude companies
        if include_companies:
            company_list = ", ".join(f"'{c}'" for c in include_companies)
            clauses.append(
                f"(company_1 IN ({company_list}) OR company_2 IN ({company_list}) OR company_3 IN ({company_list}))"
            )
        if exclude_companies:
            company_list = ", ".join(f"'{c}'" for c in exclude_companies)
            clauses.append(
                f"(company_1 NOT IN ({company_list}) AND company_2 NOT IN ({company_list}) AND company_3 NOT IN ({company_list}))"
            )

        # Include / exclude languages
        if include_languages:
            langs = ", ".join(f"'{lang}-{lang}'" for lang in include_languages)
            clauses.append(f"language IN ({langs})")
        if exclude_languages:
            langs = ", ".join(f"'{lang}-{lang}'" for lang in exclude_languages)
            clauses.append(f"language NOT IN ({langs})")

        if include_docs:
            # Assume these are URL hashes, e.g. "ENNmqkF1mGNhVhvhmbUEs4U2"
            doc_hashes = ", ".join(f"'{doc}'" for doc in include_docs)
            clauses.append(f"doc_hash IN ({doc_hashes})")

        if exclude_docs:
            # Assume these are URL hashes, e.g. "ENNmqkF1mGNhVhvhmbUEs4U2"
            doc_hashes = ", ".join(f"'{doc}'" for doc in exclude_docs)
            clauses.append(f"doc_hash NOT IN ({doc_hashes})")

        # Join everything
        if clauses:
            sql.append("WHERE " + " AND ".join(clauses))

        sql_filter = " ".join(sql)

        # Validate the SQL query against the schemas
        if not self._validate_sql(sql_filter):
            raise ValueError(f"Invalid SQL query: {sql_filter!r}. Please check your filters and try again.")

        self.logger.debug(f"Generated SQL filter: {sql_filter}")

        # Return the final SQL filter string
        return sql_filter

    def _validate_sql(self, sql: str) -> bool:
        """
        Validate a SQL query string by attempting to execute it against a mock schema.

        Parameters
        ----------
        sql : str
            The SQL query string to validate.

        Returns
        -------
        bool
            True if the SQL is valid, False otherwise.

        Examples
        --------
        >>> Nosible()._validate_sql(sql="SELECT 1")
        True
        >>> Nosible()._validate_sql(sql="SELECT * FROM missing_table")
        False
        """
        # Define a mock schema for the 'engine' table with all possible columns used in _format_sql
        columns = [
            "loc",
            "published",
            "visited",
            "certain",
            "netloc",
            "language",
            "company_1",
            "company_2",
            "company_3",
            "doc_hash",
        ]
        # Create a dummy DataFrame with correct columns and no rows
        df = pl.DataFrame({col: [] for col in columns})
        ctx = SQLContext()
        ctx.register("engine", df)
        try:
            ctx.execute(sql)
            return True
        except Exception:
            return False

    def __enter__(self) -> "Nosible":
        """
        Enter the context manager, returning this client instance.

        Returns
        -------
        Nosible
            The current client instance.
        """
        return self

    def __exit__(
        self,
        _exc_type: typing.Optional[type[BaseException]],
        _exc_val: typing.Optional[BaseException],
        _exc_tb: typing.Optional[types.TracebackType],
    ) -> typing.Optional[bool]:
        """
        Always clean up (self.close()), but let exceptions propagate.
        Return True only if you really want to suppress an exception.

        Parameters
        ----------
        exc_type : Optional[type[BaseException]]
            The type of the exception raised, if any.
        exc_val : Optional[BaseException]
            The exception instance, if any.
        exc_tb : Optional[types.TracebackType]
            The traceback object, if any.

        Returns
        -------
        Optional[bool]
            False to propagate exceptions, True to suppress them.
        """
        try:
            self.close()
        except Exception as cleanup_err:
            # optional: log or re-raise, but don’t hide the original exc
            print(f"Cleanup failed: {cleanup_err!r}")
        # Return False (or None) => exceptions inside the with‐block are re-raised.
        return False

    def __del__(self):
        """
        Destructor to ensure resources are cleaned up if not explicitly closed.

        """
        # Ensure it's called
        self.close()



import random

COMPANIES = [
    "Apple Inc.",
    "Microsoft Corporation",
    "Amazon.com, Inc.",
    "Alphabet Inc.",
    "Meta Platforms, Inc.",
    "Tesla, Inc.",
    "Berkshire Hathaway Inc.",
    "NVIDIA Corporation",
    "JPMorgan Chase & Co.",
    "Johnson & Johnson",
    "Walmart Inc.",
    "Visa Inc.",
    "Mastercard Incorporated",
    "Procter & Gamble Co.",
    "UnitedHealth Group Incorporated",
    "Bank of America Corporation",
    "Home Depot, Inc.",
    "Nestlé S.A.",
    "Samsung Electronics Co., Ltd.",
    "LVMH Moët Hennessy – Louis Vuitton",
    "ASML Holding N.V.",
    "Exxon Mobil Corporation",
    "Intel Corporation",
    "Pfizer Inc.",
    "The Coca-Cola Company",
    "PepsiCo, Inc.",
    "Chevron Corporation",
    "Merck & Co., Inc.",
    "Novartis International AG",
    "Toyota Motor Corporation",
    "Oracle Corporation",
    "Cisco Systems, Inc.",
    "Adobe Inc.",
    "Salesforce, Inc.",
    "Netflix, Inc.",
    "International Business Machines Corporation (IBM)",
    "The Walt Disney Company",
    "HSBC Holdings plc",
    "McDonald's Corporation",
    "Nike, Inc.",
    "Qualcomm Incorporated",
    "Roche Holding AG",
    "SAP SE",
    "Abbott Laboratories",
    "Costco Wholesale Corporation",
    "Broadcom Inc.",
    "Accenture plc",
    "Chevron Corporation",
    "Texas Instruments Incorporated",
    "Unilever PLC"
]

THINGS_TO_KNOW = [
    "Company name and branding",
    "Founding date and history",
    "Founders and key executives",
    "Headquarters and global offices",
    "Mission, vision, and values",
    "Core products and services",
    "Business model and revenue streams",
    "Annual revenue and growth rate",
    "Profit margins (gross, operating, net)",
    "Market capitalization and valuation",
    "Key financial ratios (P/E, ROE, ROI)",
    "Stock price history and recent performance",
    "Major investors and shareholder structure",
    "Recent mergers, acquisitions, or divestitures",
    "R&D spending and innovation pipeline",
    "Competitive landscape and main rivals",
    "Market share by region or segment",
    "Customer segments and target markets",
    "Pricing strategy and positioning",
    "Supply chain structure and partners",
    "Distribution channels and logistics",
    "Marketing and advertising strategies",
    "Brand perception and reputation",
    "ESG (Environmental, Social, Governance) scores",
    "Sustainability initiatives and impact",
    "Corporate culture and employee count",
    "Employee satisfaction and turnover rates",
    "Leadership and governance practices",
    "Regulatory environment and compliance record",
    "Key risks and litigation history",
    "Recent news, press releases, and media coverage",
    "Patents, trademarks, and IP portfolio",
    "Digital transformation and tech stack",
    "Website traffic and social media metrics",
    "Mobile app usage and customer reviews",
    "Partnerships and strategic alliances",
    "Future outlook and analyst recommendations",
]

JOB_TITLES = [
    "Chief Executive Officer (CEO)",
    "Chief Financial Officer (CFO)",
    "Chief Operating Officer (COO)",
    "Chief Technology Officer (CTO)",
    "Chief Marketing Officer (CMO)",
    "Head of Investor Relations",
    "Director of Corporate Strategy",
    "Business Development Manager",
    "Product Manager",
    "Marketing Manager",
    "Brand Manager",
    "Financial Analyst",
    "Equity Research Analyst",
    "Market Research Analyst",
    "Consultant",
    "Venture Capital Associate",
    "Private Equity Associate",
    "Operations Manager",
    "Supply Chain Manager",
    "Human Resources Manager",
    "Sustainability Officer",
    "Compliance Officer",
    "Legal Counsel",
    "Risk Manager",
    "Data Analyst",
    "IT Manager",
    "Sales Director",
    "Account Manager",
    "Customer Success Manager",
    "Public Relations Manager"
]

def _get_question():
    return (f"I am a {random.choice(JOB_TITLES)} and I want to know {random.choice(THINGS_TO_KNOW)}"
            f"about {random.choice(COMPANIES)}")


from dataclasses import asdict, dataclass, field

from nosible.utils.json_tools import json_dumps, print_dict


@dataclass(init=True, repr=True, eq=True, frozen=True)
class Snippet:
    """
    A class representing a snippet of text, typically extracted from a web page.

    Parameters
    ----------
    content : str or None
        The text content of the snippet.
    images : list or None
        List of image URLs associated with the snippet.
    language : str or None
        The language of the snippet.
    next_snippet_hash : str or None
        Hash of the next snippet in sequence.
    prev_snippet_hash : str or None
        Hash of the previous snippet in sequence.
    snippet_hash : str or None
        A unique hash for the snippet.
    statistics : dict or None
        Statistical information about the snippet.
    url_hash : str or None
        Hash of the URL from which the snippet was extracted.
    words : str or None
        The words in the snippet.
    links : list or None
        List of links associated with the snippet.
    companies : list or None
        List of companies mentioned in the snippet.


    Examples
    --------
    >>> snippet = Snippet(content="Example snippet", language="en")
    >>> print(snippet.content)
    Example snippet

    """

    content: str = field(default=None, repr=True, compare=True)
    """The text content of the snippet."""
    images: list = field(default=None, repr=True, compare=False)
    """List of image URLs associated with the snippet."""
    language: str = field(default=None, repr=True, compare=False)
    """The language of the snippet."""
    next_snippet_hash: str = field(default=None, repr=True, compare=False)
    """Hash of the next snippet in sequence."""
    prev_snippet_hash: str = field(default=None, repr=True, compare=False)
    """Hash of the previous snippet in sequence."""
    snippet_hash: str = field(default=None, repr=True, compare=True)
    """A unique hash for the snippet."""
    statistics: dict = field(default=None, repr=False, compare=False)
    """Statistical information about the snippet."""
    url_hash: str = field(default=None, repr=True, compare=False)
    """Hash of the URL from which the snippet was extracted."""
    words: str = field(default=None, repr=False, compare=False)
    """The words in the snippet."""
    links: list = field(default=None, repr=False, compare=False)
    """List of links associated with the snippet."""
    companies: list = field(default=None, repr=False, compare=False)
    """List of companies mentioned in the snippet."""

    def __str__(self):
        """
        Returns a user-friendly string representation of the Snippet.

        Returns
        -------
        str
            A string representation of the Snippet.
        """
        return print_dict(self.to_dict())

    def __getitem__(self, key: str):
        """
        Allows access to snippet attributes using dictionary-like syntax.

        Parameters
        ----------
        key : str
            The attribute name to access.

        Returns
        -------
        Any
            The value of the specified attribute.

        Raises
        ------
        KeyError
            If the key does not match any attribute.
        """
        if hasattr(self, key):
            return getattr(self, key)
        raise KeyError(f"'{key}' is not a valid Snippet attribute.")

    def to_dict(self) -> dict:
        """
        Convert the Snippet to a dictionary representation.

        Returns
        -------
        dict
            Dictionary containing all fields of the Snippet.

        Examples
        --------
        >>> snippet = Snippet(content="Example snippet", snippet_hash="hash1")
        >>> snippet_dict = snippet.to_dict()
        >>> isinstance(snippet_dict, dict)
        True
        """
        return asdict(self, dict_factory=dict)

    @classmethod
    def from_dict(cls, data: dict) -> "Snippet":
        """
        Create a Snippet instance from a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary containing snippet data.

        Returns
        -------
        Snippet
            An instance of Snippet populated with the provided data.

        Examples
        --------
        >>> snippet_data = {"content": "Example snippet", "snippet_hash": "hash1"}
        >>> snippet = Snippet.from_dict(snippet_data)
        >>> isinstance(snippet, Snippet)
        True
        """
        return cls(**data)

    def to_json(self) -> str:
        """
        Convert the Snippet to a JSON string representation.

        Returns
        -------
        str
            JSON string containing all fields of the Snippet.

        Examples
        --------
        >>> snippet = Snippet(content="Example snippet", snippet_hash="hash1")
        >>> json_str = snippet.to_json()
        >>> isinstance(json_str, str)
        True
        """
        return json_dumps(self.to_dict())



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field

import duckdb
import pandas as pd
import polars as pl
from tantivy import Document, Index, SchemaBuilder

from nosible.classes.result import Result
from nosible.utils.json_tools import json_dumps, json_loads


@dataclass(frozen=True)
class ResultSet(Iterator[Result]):
    """
    Container class for managing and processing a sequence of Result objects.

    This class provides methods for iterating, accessing, and converting collections of
    Result instances. It supports context management, sequence operations, and
    conversion to and from various data formats (CSV, JSON, DataFrames, etc.).

    Parameters
    ----------
    results : list of Result
        The list of Result objects contained in the ResultSet.

    Examples
    --------
    >>> from nosible import Result, ResultSet
    >>> results = [
    ...     Result(url="https://example.com", title="Example Domain"),
    ...     Result(url="https://openai.com", title="OpenAI"),
    ... ]
    >>> search_results = ResultSet(results)
    >>> len(search_results)
    2
    >>> for result in search_results:
    ...     print(result.title)
    Example Domain
    OpenAI
    >>> df = search_results.to_pandas()
    >>> list(df.columns)  # doctest: +ELLIPSIS
    ['url', 'title', 'description', 'netloc', 'published', 'visited', 'author', 'content', ... 'url_hash']
    """

    _FIELDS = [
        "url",
        "title",
        "description",
        "netloc",
        "published",
        "visited",
        "author",
        "content",
        "language",
        "similarity",
        "url_hash",
    ]

    results: list[Result] = field(default_factory=list)
    """ List of Result objects contained in this ResultSet."""
    _index: int = field(default=0, init=False, repr=False, compare=False)
    """ Internal index for iteration over results."""

    def __len__(self) -> int:
        """
        Return the number of search results.

        Returns
        -------
        int
            The number of Result objects in the results list.
        """
        return len(self.results)

    def __str__(self) -> str:
        """
        Return a string representation of the ResultSet.

        Returns
        -------
        str
            A formatted table showing the index, similarity, and title of each Result.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain", similarity=0.95),
        ...     Result(url="https://openai.com", title="OpenAI", similarity=0.99),
        ... ]
        >>> search_results = ResultSet(results)
        >>> print(search_results)  # doctest: +NORMALIZE_WHITESPACE
        Idx | Similarity | Title
        ------------------------
          0 |   0.95     | Example Domain
          1 |   0.99     | OpenAI

        >>> empty = ResultSet([])
        >>> print(empty)
        ResultSet: No results found.
        """
        if not self.results:
            return "ResultSet: No results found."

        # Create a formatted string for each result
        lines = []
        for idx, result in enumerate(self.results):
            similarity = f"{result.similarity:.2f}" if result.similarity is not None else "  N/A"
            title = result.title or "No Title"
            lines.append(f"{idx:>3} | {similarity:>10} | {title}")

        # Add a header with matching column widths
        header = f"{'Idx':>3} | {'Similarity':>10} | Title"
        separator = "-" * len(header)
        lines.insert(0, header)
        lines.insert(1, separator)
        # Join all lines into a single string
        return "\n".join(lines)

    def __iter__(self) -> ResultSet:
        """
        Reset iteration and return self.

        Returns
        -------
        ResultSet
            Iterator over the ResultSet instance.
        """
        object.__setattr__(self, "_index", 0)
        return self

    def __next__(self) -> Result:
        """
        Returns the next Result in the sequence.

        Returns
        -------
        Result
            The next Result object in the sequence.
        Raises
        ------
        StopIteration
            If the end of the sequence is reached.
        """
        if self._index < len(self.results):
            item = self.results[self._index]
            object.__setattr__(self, "_index", self._index + 1)
            return item
        raise StopIteration

    def __eq__(self, value):
        """
        Comapre set of url_hashes to determine equality.
        Two ResultSet instances are considered equal if they contain the same set of url_hashes.

        Parameters
        ----------
        value : ResultSet
            The ResultSet instance to compare against.
        Returns
        -------
        bool
            True if both ResultSet instances contain the same set of url_hashes, False otherwise.
        """
        if not isinstance(value, ResultSet):
            return False
        # Compare the sets of url_hashes
        return {r.url_hash for r in self.results} == {r.url_hash for r in value.results}

    def __enter__(self) -> ResultSet:
        """
        Enters the runtime context related to this object.

        Returns
        -------
        ResultSet
            The context manager instance itself.
        """
        # Setup if required
        return self

    def __getitem__(self, key: int) -> Result:
        """
        Get a Result by index.

        Parameters
        ----------
        key : int
            Index of the result to retrieve.

        Returns
        -------
        Result
            The Result at the specified index.

        Raises
        ------
        IndexError
            If index is out of range.
        """
        if 0 <= key < len(self.results):
            return self.results[key]
        raise IndexError(f"Index {key} out of range for ResultSet with length {len(self.results)}.")
        raise IndexError(f"Index {key} out of range for ResultSet with length {len(self.results)}.")

    def __add__(self, other: ResultSet | Result) -> ResultSet:
        """
        Concatenate two ResultSet instances.

        Parameters
        ----------
        other : ResultSet
            Another ResultSet instance to concatenate with.

        Returns
        -------
        ResultSet
            A new ResultSet instance containing results from both instances.

        Raises
        ------
        TypeError

        Examples
        --------
        >>> results1 = ResultSet([Result(url="https://example.com")])
        >>> results2 = ResultSet([Result(url="https://openai.com")])
        >>> combined = results1 + results2
        >>> len(combined)
        2
        """
        if isinstance(other, ResultSet):
            return ResultSet(self.results + other.results)
        if isinstance(other, Result):
            # If other is a single Result, create a new ResultSet with it
            return ResultSet(self.results.append(other))
        raise TypeError("Can only concatenate ResultSet with another ResultSet.")

    def __sub__(self, other: ResultSet) -> ResultSet:
        """
        Subtract another ResultSet from this one, returning a new ResultSet with results not in the other.

        Parameters
        ----------
        other : ResultSet
            Another ResultSet instance to subtract from this one.

        Returns
        -------
        ResultSet
            A new ResultSet instance containing results from this instance that are not in the other.

        Raises
        ------
        TypeError


        Examples
        --------
        >>> results1 = ResultSet([Result(url="https://example.com")])
        >>> results2 = ResultSet([Result(url="https://openai.com")])
        >>> difference = results1 - results2
        >>> len(difference)
        1
        """
        if not isinstance(other, ResultSet):
            raise TypeError("Can only subtract ResultSet with another ResultSet.")
        # Use url_hash for set-based difference if available, else fallback to object identity
        other_hashes = {r.url_hash for r in other.results if r.url_hash}
        if other_hashes:
            filtered = [r for r in self.results if r.url_hash not in other_hashes]
        else:
            # Fallback: use object identity (slower)
            filtered = [r for r in self.results if r not in other.results]
        return ResultSet(filtered)

    def __del__(self) -> None:
        """
        Cleanup resources when the ResultSet instance is deleted. TODO
        """
        self.close()

    def find_in_search_results(self, query: str, top_k: int = 10) -> ResultSet:
        """
        Perform an in-memory search over a ResultSet collection using Tantivy.

        Parameters
        ----------
        query : str
            The search string to rank within these results.
        top_k : int
            Number of top results to return.

        Returns
        -------
        ResultSet
            A new ResultSet instance containing the top_k results ranked by relevance to `query`.

        Examples
        --------
        >>> from nosible import Nosible
        >>> from nosible import ResultSet
        >>> with Nosible() as nos:
        ...     results: ResultSet = nos.search(question="Aircraft Manufacturing", n_results=10)
        ...     inner = results.find_in_search_results("embraer", top_k=5)
        >>> print(f"Top {len(inner)} hits for “embraer” within the initial results:")
        Top 5 hits for “embraer” within the initial results:
        >>> for idx, hit in enumerate(inner, start=1):
        ...     print("Document returned")
        Document returned
        Document returned
        Document returned
        Document returned
        Document returned
        """
        # Build the Tantivy schema
        schema_builder = SchemaBuilder()
        # Int for doc retrieval.
        schema_builder.add_integer_field("doc_id", stored=True)
        # Content will hold the concatenated text you want to search.
        schema_builder.add_text_field("content", stored=True)
        schema = schema_builder.build()

        # Create in-memory index and writer.
        index = Index(schema)

        # 15MB of RAM reserved (minimum you can).
        writer = index.writer(heap_size=15_000_000, num_threads=1)

        # Index each Result
        for idx, result in enumerate(self.results):
            parts = [result.title or "", result.description or "", result.content or ""]
            full_text = " ".join(p for p in parts if p)

            doc = Document()
            # Pass field names as strings.
            doc.add_integer("doc_id", idx)
            doc.add_text("content", full_text)
            writer.add_document(doc)
        # Flushes and writes the inverted index.
        writer.commit()
        # Makes that new data visible to searchers.
        index.reload()

        # Search the in-memory index
        searcher = index.searcher()
        tantivy_query = index.parse_query(query, ["content"])

        # Map Tantivy hits back to original indices
        hits = searcher.search(tantivy_query, top_k).hits
        matched_idxs = [searcher.doc(addr).get_first("doc_id") for (_score, addr) in hits]
        top_results = [self.results[i] for i in matched_idxs]

        # Pad out to top_k with the remaining docs, in original order.
        if len(top_results) < top_k:
            for i, doc in enumerate(self.results):
                if i not in matched_idxs:
                    top_results.append(doc)
                    if len(top_results) == top_k:
                        break

        return ResultSet(top_results)

    def analyze(self, by: str = "published") -> dict:
        """
        Analyze ResultSet by grouping on a specified field.

        This method uses Polars to compute different metrics based on the `by` parameter:

        - **Date fields** (`"published"` or `"visited"`): Counts per month.
        - **Similarity**: Descriptive statistics.
        - **Categorical fields**: value counts sorted descending, mapping each value to its frequency.

        Parameters
        ----------
        by : str
            The Result attribute to analyze. Must be one of:
            'netloc', 'published', 'visited', 'author', 'language',
             or 'similarity'

        Raises
        ------
        ValueError

        Returns
        -------
        dict
            A dictionary of analysis results. The structure depends on `by`:

            - Date-based fields: { 'YYYY-MM': int(count), ... }
            - Numeric fields: { 'count': float, 'mean': float, 'std': float,
                                'min': float, '25%': float, '50%': float,
                                '75%': float, 'max': float }
            - Categorical fields: { value: int(count), ... } sorted by count descending.

        Examples
        --------
        >>> from nosible import Nosible
        >>> from nosible import Result, ResultSet
        >>> with Nosible() as nos:
        ...     results: ResultSet = nos.search(question="Aircraft Manufacturing", n_results=100)
        ...     summary = results.analyze(by="language")
        ...     print(summary)
        {'en': 100}
        >>> import polars as pl
        >>> from nosible.classes.result_set import Result, ResultSet

        # -- date grouping (published) --------------------------------------------
        >>> data = [
        ...     {"published": "2021-01-15", "netloc": "a.com", "author": "", "language": "en", "similarity": 0.5},
        ...     {"published": "2021-02-20", "netloc": "a.com", "author": "", "language": "en", "similarity": 0.8},
        ...     {"published": "2021-02-25", "netloc": "b.org", "author": "", "language": "fr", "similarity": 0.2},
        ... ]
        >>> results = ResultSet([Result(**d) for d in data])
        >>> results.analyze(by="published")  # doctest: +NORMALIZE_WHITESPACE
        {'2021-01': 1, '2021-02': 2}

        # -- numeric stats (similarity) ------------------------------------------
        >>> stats = results.analyze(by="similarity")
        >>> set(stats) == {"count", "null_count", "mean", "std", "min", "25%", "50%", "75%", "max"}
        True
        >>> round(stats["mean"], 2)
        0.5

        # -- categorical counts (language) --------------------------------------
        >>> results.analyze(by="language")
        {'en': 2, 'fr': 1}

        # -- author special case ------------------------------------------------
        # empty author strings get mapped to "Author Unknown"
        >>> results.analyze(by="author")
        {'Author Unknown': 3}

        # -- invalid field -------------------------------------------------------
        >>> results.analyze(by="foobar")  # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ValueError: Cannot analyze by 'foobar' - not a valid field.
        """
        # Convert to Polars DataFrame
        df: pl.DataFrame = self.to_polars()

        # Validate column
        if by not in ["netloc", "published", "visited", "author", "language", "similarity"]:
            raise ValueError(f"Cannot analyze by '{by}' - not a valid field.")

        # Drop nulls for the analysis column
        df = df.drop_nulls(by)
        if df.is_empty():
            return {}

        # Handle author unknown
        if by == "author":
            df = df.with_columns(
                pl.when(pl.col("author") == "")
                .then(pl.lit("Author Unknown"))
                .otherwise(pl.col("author"))
                .alias("author")
            )

        # Handle date fields: 'published' or 'visited'
        if by in ("published", "visited"):
            # parse ISO date strings to Date
            df = df.with_columns(pl.col(by).str.strptime(pl.Date, "%Y-%m-%d", strict=False).alias(by))
            # Extract year-month
            df = df.with_columns(pl.col(by).dt.strftime("%Y-%m").alias("year_month"))
            # Count per month
            vc = df.group_by("year_month").agg(pl.count().alias("count")).sort("year_month")
            rows = vc.rows()
            if not rows:
                return {}
            # Generate full month range
            first_month, last_month = rows[0][0], rows[-1][0]
            all_months = pd.date_range(start=f"{first_month}-01", end=f"{last_month}-01", freq="MS").strftime("%Y-%m")
            result = dict.fromkeys(all_months, 0)
            # Fill actual counts
            for month, cnt in rows:
                result[month] = cnt
            return result

        # Numeric stats for similarity
        if by == "similarity":
            desc_df = df["similarity"].describe()
            # print({row[0]: float(row[1]) for row in desc_df.rows()})
            return {row[0]: float(row[1]) for row in desc_df.rows()}

        # Non-date: analyze numeric vs. categorical Non-date: analyze numeric vs. categorical
        series = df[by]

        # Categorical/value counts
        vc = series.value_counts()
        _, count_col = vc.columns
        sorted_vc = vc.sort(count_col, descending=True)
        return {str(row[0]): int(row[1]) for row in sorted_vc.rows()}

    # Conversion methods
    def to_csv(self, file_path: str | None = None, delimiter: str = ",", encoding: str = "utf-8") -> str:
        """
        Serialize the search results to a CSV file.

        This method writes the current ResultSet to a CSV file using Python's built-in
        csv module. Each Result is converted to a dictionary and written as a row.
        The CSV will contain all fields defined in the FIELDS class attribute.

        Parameters
        ----------
        file_path : str or None, optional
            Path to save the CSV file.
        delimiter : str, optional
            Delimiter to use in the CSV file.
        encoding : str, optional
            Encoding for the CSV file.

        Returns
        -------
        str
            The path to the written CSV file.

        Raises
        ------
        RuntimeError
            If writing to the CSV file fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> path = search_results.to_csv("out.csv")
        >>> path.endswith(".csv")
        True
        """
        import csv

        out = file_path or "search_results.csv"
        try:
            with open(out, "w", newline="", encoding=encoding) as f:
                writer = csv.DictWriter(f, fieldnames=self._FIELDS, delimiter=delimiter)
                writer.writeheader()
                for result in self.results:
                    writer.writerow(result.to_dict())
        except Exception as e:
            raise RuntimeError(f"Failed to write CSV to '{out}': {e}") from e
        return out

    def to_polars(self) -> pl.DataFrame:
        """
        Convert the search results to a Polars DataFrame.

        Returns
        -------
        pl.DataFrame
            A Polars DataFrame containing all search results, with columns for each field.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> df = search_results.to_polars()
        >>> isinstance(df, pl.DataFrame)
        True
        >>> "url" in df.columns
        True
        """
        return pl.DataFrame(self.to_dicts())

    def to_pandas(self) -> pd.DataFrame:
        """
        Convert the search results to a pandas DataFrame.

        Returns
        -------
        pandas.DataFrame
            DataFrame containing all search results, with columns for each field.

        Raises
        ------
        RuntimeError
            If conversion to Pandas DataFrame fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> df = search_results.to_pandas()
        >>> import pandas as pd
        >>> isinstance(df, pd.DataFrame)
        True
        >>> {"url", "title"}.issubset(df.columns)
        True
        """
        try:
            return self.to_polars().to_pandas()
        except Exception as e:
            raise RuntimeError(f"Failed to convert search results to Pandas DataFrame: {e}") from e

    def to_json(self, file_path: str | None = None) -> str | bytes:
        """
        Serialize the search results to a JSON string and optionally write to disk.

        Parameters
        ----------
        file_path : str or None, optional
            Path to save the JSON file. If None, the JSON string is returned.

        Returns
        -------
        str
            The JSON string if `file_path` is None, otherwise the JSON string that was written to file.
        Raises
        -------
        RuntimeError
            If serialization to JSON fails or if writing to the file fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> json_str = search_results.to_json()
        >>> isinstance(json_str, str)
        True
        >>> # Optionally write to file
        >>> path = search_results.to_json(file_path="results.json")
        >>> path.endswith(".json")
        True
        """
        try:
            json_bytes = json_dumps(self.to_dicts())
            if file_path:
                try:
                    with open(file_path, "w") as f:
                        f.write(json_bytes)
                    return file_path
                except Exception as e:
                    raise RuntimeError(f"Failed to write JSON to '{file_path}': {e}") from e
            return json_bytes
        except Exception as e:
            raise RuntimeError(f"Failed to serialize results to JSON: {e}") from e

    def to_dicts(self) -> list[dict]:
        """
        Return the search results as a list of dictionaries.

        Each Result in the collection is converted to a dictionary
        containing all its fields. This is useful for serialization,
        inspection, or further processing.

        Returns
        -------
        list of dict
            List where each element is a dictionary representation of a Result.

        Raises
        ------
        RuntimeError
            If conversion to list of dictionaries fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> dict_list = search_results.to_dicts()
        >>> isinstance(dict_list, list)
        True
        >>> isinstance(dict_list[0], dict)
        True
        >>> "url" in dict_list[0]
        True
        """
        try:
            return [result.to_dict() for result in self.results]
        except Exception as e:
            raise RuntimeError(f"Failed to convert results to list of dictionaries: {e}") from e

    def to_dict(self) -> dict:
        """
        Return the search results as a dictionary keyed by `url_hash`.

        Each entry in the returned dictionary maps a unique `url_hash` to the
        corresponding search result's dictionary representation. Only results
        with a non-empty `url_hash` are included.

        Returns
        -------
        dict
            Dictionary mapping `url_hash` (str) to the result dictionary.

        Raises
        ------
        RuntimeError
            If conversion to dictionary fails, e.g., if any Result lacks a `url_hash`.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", url_hash="abc123", title="Example Domain"),
        ...     Result(url="https://openai.com", url_hash="def456", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> result_dict = search_results.to_dict()
        >>> isinstance(result_dict, dict)
        True
        >>> list(result_dict.keys())
        ['abc123', 'def456']
        >>> result_dict["abc123"]["title"]
        'Example Domain'
        """
        try:
            return {r.url_hash: r.to_dict() for r in self.results if r.url_hash}
        except Exception as e:
            raise RuntimeError(f"Failed to convert results to dict: {e}") from e

    def to_ndjson(self, file_path: str | None = None) -> str:
        """
        Serialize search results to newline-delimited JSON (NDJSON) format.

        Each search result is serialized as a single JSON object per line.
        The resulting NDJSON string can be written to disk or returned as a string.

        Parameters
        ----------
        file_path : str or None, optional
            Path to save the NDJSON file. If None, returns the NDJSON string.

        Returns
        -------
        str
            File path if written to file, otherwise the NDJSON string.

        Raises
        ------
        RuntimeError
            If serialization to NDJSON fails or if writing to the file fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> ndjson_str = search_results.to_ndjson()
        >>> print(ndjson_str.splitlines()[0])  # doctest: +ELLIPSIS
        {"url":"https://example.com","title":"Example Domain","description":null,"netloc":null..."url_hash":null}
        >>> # Optionally write to file
        >>> path = search_results.to_ndjson(file_path="results.ndjson")
        >>> path.endswith(".ndjson")
        True
        """

        ndjson_lines = []
        for result in self.results:
            try:
                ndjson_lines.append(json_dumps(result.to_dict()))
            except Exception as e:
                raise RuntimeError(f"Failed to serialize Result to NDJSON: {e}") from e

        if file_path:
            try:
                with open(file_path, "w", encoding="utf-8") as f:
                    f.write("\n".join(ndjson_lines) + "\n")
                return file_path
            except Exception as e:
                raise RuntimeError(f"Failed to write NDJSON to '{file_path}': {e}") from e
        return "\n".join(ndjson_lines) + "\n"

    def to_parquet(self, file_path: str | None = None) -> str:
        """
        Serialize the search results to Apache Parquet format using Polars.

        This method writes the current ResultSet to a Parquet file, which is an efficient
        columnar storage format suitable for analytics and interoperability with data tools.

        Parameters
        ----------
        file_path : str or None, optional
            Path to save the Parquet file.

        Returns
        -------
        str
            The path to the written Parquet file.

        Raises
        ------
        RuntimeError
            If writing to the Parquet file fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> parquet_path = search_results.to_parquet("my_results.parquet")
        >>> parquet_path.endswith(".parquet")
        True
        """
        out = file_path or "results.parquet"
        try:
            self.to_polars().write_parquet(out)
        except Exception as e:
            raise RuntimeError(f"Failed to write Parquet to '{out}': {e}") from e
        return out

    def to_arrow(self, file_path: str | None = None) -> str:
        """
        Serialize the search results to Apache Arrow IPC (Feather) format using Polars.

        This method writes the current ResultSet to an Arrow IPC file, which is an efficient
        columnar storage format for fast data interchange and analytics.

        Parameters
        ----------
        file_path : str or None, optional
            Path to save the Arrow IPC file.

        Returns
        -------
        str
            The path to the written Arrow IPC file.

        Raises
        ------
        RuntimeError
            If writing to the Arrow IPC file fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> arrow_path = search_results.to_arrow("my_results.arrow")
        >>> arrow_path.endswith(".arrow")
        True
        """
        out = file_path or "results.arrow"
        try:
            self.to_polars().write_ipc(out)
        except Exception as e:
            raise RuntimeError(f"Failed to write Arrow IPC to '{out}': {e}") from e
        return out

    def to_duckdb(self, file_path: str | None = None, table_name: str = "results") -> str:
        """
        Serialize the search results to a DuckDB database file and table.

        This method writes the current ResultSet to a DuckDB database file,
        creating a new table or replacing an existing one with the specified name.
        The table will contain all fields defined in the FIELDS class attribute.

        Parameters
        ----------
        file_path : str or None, optional
            Path to save the DuckDB file.
        table_name : str, optional
            Name of the table to write the results to.

        Returns
        -------
        str
            The path to the written DuckDB file.

        Raises
        ------
        RuntimeError
            If writing to the DuckDB file fails.

        Examples
        --------
        >>> from nosible import Result, ResultSet
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain"),
        ...     Result(url="https://openai.com", title="OpenAI"),
        ... ]
        >>> search_results = ResultSet(results)
        >>> db_path = search_results.to_duckdb(file_path="my_results.duckdb", table_name="search_table")
        >>> db_path.endswith(".duckdb")
        True
        """
        out = file_path or "results.duckdb"
        try:
            import duckdb

            # Convert to Polars DataFrame and then to Arrow Table
            df = self.to_polars()
            # Connect to DuckDB and write the Arrow Table to a table
            con = duckdb.connect(out)
            # Write the DataFrame to the specified table name, replacing if exists
            con.execute(f"CREATE OR REPLACE TABLE {table_name} AS SELECT * FROM df").df()
            con.close()
            return out

        except Exception as e:
            raise RuntimeError(f"Failed to write DuckDB table to '{out}': {e}") from e

    # Loading from disk
    @classmethod
    def from_csv(cls, file_path: str) -> ResultSet:
        """
        Load search results from a CSV file using Polars.

        Parameters
        ----------
        file_path : str
            Path to the CSV file.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        RuntimeError
            If reading the CSV file fails or if parsing rows fails.
        ValueError
            If a row in the CSV does not match the expected format.

        Examples
        --------
        >>> import polars as pl
        >>> from nosible import ResultSet, Result
        >>> # Suppose 'data.csv' contains columns: url,title,description
        >>> _ = pl.DataFrame(
        ...     [
        ...         {"url": "https://example.com", "title": "Example Domain", "description": "Example description"},
        ...         {"url": "https://openai.com", "title": "OpenAI", "description": "AI research"},
        ...     ]
        ... ).write_csv("data.csv")
        >>> results = ResultSet.from_csv("data.csv")
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        try:
            df = pl.read_csv(file_path)
        except Exception as e:
            raise RuntimeError(f"Failed to read CSV file '{file_path}': {e}") from e
        results = []
        for row in df.iter_rows(named=True):
            try:
                result = Result(
                    url=row.get("url"),
                    title=row.get("title"),
                    description=row.get("description"),
                    netloc=row.get("netloc"),
                    published=row.get("published"),
                    visited=row.get("visited"),
                    author=row.get("author"),
                    content=row.get("content"),
                    language=row.get("language"),
                    similarity=row.get("similarity"),
                    url_hash=row.get("url_hash"),
                )
                results.append(result)
            except Exception as e:
                raise ValueError(f"Error parsing row in CSV: {row}\n{e}") from e
        return cls(results)

    @classmethod
    def from_json(cls, file_path: str) -> ResultSet:
        """
        Load search results from a JSON file.

        Parameters
        ----------
        file_path : str
            Path to the JSON file containing search results.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        RuntimeError
            If reading the JSON file fails or if parsing the data fails.


        Examples
        --------
        >>> import json
        >>> from nosible import ResultSet
        >>> with open("data.json", "w") as f:
        ...     json.dump(
        ...         [
        ...             {"url": "https://example.com", "title": "Example Domain"},
        ...             {"url": "https://openai.com", "title": "OpenAI"},
        ...         ],
        ...         f,
        ...     )
        >>> results = ResultSet.from_json("data.json")
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        try:
            with open(file_path) as f:
                data = json_loads(f.read())
        except Exception as e:
            raise RuntimeError(f"Failed to read or parse JSON file '{file_path}': {e}") from e
        try:
            return cls.from_dict(data)
        except Exception as e:
            raise RuntimeError(f"Failed to create ResultSet from JSON data in '{file_path}': {e}") from e

    @classmethod
    def from_polars(cls, df: pl.DataFrame) -> ResultSet:
        """
        Create a ResultSet instance from a Polars DataFrame.

        Parameters
        ----------
        df : pl.DataFrame
            Polars DataFrame containing columns corresponding to Result fields.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        ValueError
            If a row in the DataFrame does not match the expected format or if required fields are

        Examples
        --------
        >>> import polars as pl
        >>> from nosible import ResultSet
        >>> data = [
        ...     {"url": "https://example.com", "title": "Example Domain", "similarity": 0.95},
        ...     {"url": "https://openai.com", "title": "OpenAI", "similarity": 0.99},
        ... ]
        >>> df = pl.DataFrame(data)
        >>> results = ResultSet.from_polars(df)
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        results = []
        for row in df.iter_rows(named=True):
            try:
                result = Result(
                    url=row.get("url"),
                    title=row.get("title"),
                    description=row.get("description"),
                    netloc=row.get("netloc"),
                    published=row.get("published"),
                    visited=row.get("visited"),
                    author=row.get("author"),
                    content=row.get("content"),
                    language=row.get("language"),
                    similarity=row.get("semantics", {}).get("similarity", row.get("similarity")),
                    url_hash=row.get("url_hash"),
                )
                results.append(result)
            except Exception as e:
                raise ValueError(f"Error parsing row in DataFrame: {row}\n{e}") from e
        return cls(results)

    @classmethod
    def from_pandas(cls, df: pd.DataFrame) -> ResultSet:
        """
        Create a ResultSet instance from a pandas DataFrame.
        This class method converts a given pandas DataFrame to a Polars DataFrame
        and then constructs a ResultSet object from it. This is useful for
        integrating with workflows that use pandas for data manipulation.

        Parameters
        ----------
        df : pandas.DataFrame
            DataFrame containing the search result fields. Each row should represent a single search result, with
            columns corresponding to the expected fields of ResultSet.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing the data from the input DataFrame.

        Examples
        --------
        >>> data = [{"url": "https://example.com", "title": "Example"}]
        >>> df = pd.DataFrame(data)
        >>> print(len(df))
        1
        """
        pl_df = pl.from_pandas(df)
        return cls.from_polars(pl_df)

    @classmethod
    def from_ndjson(cls, file_path: str) -> ResultSet:
        """
        Load search results from a newline-delimited JSON (NDJSON) file.

        Parameters
        ----------
        file_path : str
            Path to the NDJSON file containing search results, where each line is a JSON object.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        RuntimeError
            If the file cannot be read.
        ValueError
            If no valid search results are found or a line cannot be parsed.

        Examples
        --------
        Suppose 'data.ndjson' contains:
            {"url": "https://example.com", "title": "Example Domain"}
            {"url": "https://openai.com", "title": "OpenAI"}

        >>> from nosible import ResultSet, Result
        >>> # Write example NDJSON file
        >>> with open("data.ndjson", "w") as f:
        ...     f.write('{"url": "https://example.com", "title": "Example Domain"}\\n')
        ...     f.write('{"url": "https://openai.com", "title": "OpenAI"}\\n')
        58
        49
        >>> results = ResultSet.from_ndjson("data.ndjson")
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        results = []
        try:
            with open(file_path) as f:
                for line in f:
                    if line.strip():
                        try:
                            data = json_loads(line)
                            result = Result(
                                url=data.get("url"),
                                title=data.get("title"),
                                description=data.get("description"),
                                netloc=data.get("netloc"),
                                published=data.get("published"),
                                visited=data.get("visited"),
                                author=data.get("author"),
                                content=data.get("content"),
                                language=data.get("language"),
                                similarity=data.get("similarity"),
                                url_hash=data.get("url_hash"),
                            )
                            results.append(result)
                        except Exception as e:
                            raise ValueError(f"Error parsing NDJSON line: {line}\n{e}") from e
        except Exception as e:
            raise RuntimeError(f"Failed to read NDJSON file '{file_path}': {e}") from e
        if not results:
            raise ValueError(f"No valid search results found in the NDJSON file '{file_path}'.")
        return cls(results)

    @classmethod
    def from_parquet(cls, file_path: str) -> ResultSet:
        """
        Load search results from a Parquet file using Polars.

        Parameters
        ----------
        file_path : str
            Path to the Parquet file containing search results.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        RuntimeError
            If the file cannot be read or parsed.

        Examples
        --------
        >>> import polars as pl
        >>> from nosible import ResultSet
        >>> # Create a sample Parquet file
        >>> df = pl.DataFrame(
        ...     [
        ...         {"url": "https://example.com", "title": "Example Domain", "similarity": 0.95},
        ...         {"url": "https://openai.com", "title": "OpenAI", "similarity": 0.99},
        ...     ]
        ... )
        >>> df.write_parquet("sample.parquet")
        >>> results = ResultSet.from_parquet("sample.parquet")
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        try:
            df = pl.read_parquet(file_path)
        except Exception as e:
            raise RuntimeError(f"Failed to read Parquet file '{file_path}': {e}") from e
        try:
            return cls.from_polars(df)
        except Exception as e:
            raise RuntimeError(f"Failed to create ResultSet from Parquet data in '{file_path}': {e}") from e

    @classmethod
    def from_arrow(cls, file_path: str) -> ResultSet:
        """
        Load search results from an Apache Arrow IPC (Feather) file using Polars.

        Parameters
        ----------
        file_path : str
            Path to the Arrow IPC file containing search results.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        RuntimeError
            If the file cannot be read or parsed.

        Examples
        --------
        >>> import polars as pl
        >>> from nosible import ResultSet
        >>> # Create a sample Arrow IPC file
        >>> df = pl.DataFrame(
        ...     [
        ...         {"url": "https://example.com", "title": "Example Domain", "similarity": 0.95},
        ...         {"url": "https://openai.com", "title": "OpenAI", "similarity": 0.99},
        ...     ]
        ... )
        >>> df.write_ipc("sample.arrow")
        >>> results = ResultSet.from_arrow("sample.arrow")
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        try:
            df = pl.read_ipc(file_path)
        except Exception as e:
            raise RuntimeError(f"Failed to read Arrow IPC file '{file_path}': {e}") from e
        try:
            return cls.from_polars(df)
        except Exception as e:
            raise RuntimeError(f"Failed to create ResultSet from Arrow data in '{file_path}': {e}") from e

    @classmethod
    def from_duckdb(cls, file_path: str) -> ResultSet:
        """
        Load search results from a DuckDB database file.

        This class method reads a DuckDB file, retrieves the first available table,
        and loads its contents as Result objects. The table is expected to have
        columns matching the Result fields.

        Parameters
        ----------
        file_path : str
            Path to the DuckDB database file.

        Returns
        -------
        ResultSet
            An instance of ResultSet containing all loaded results.

        Raises
        ------
        RuntimeError
            If the file cannot be read or parsed.
        ValueError
            If no tables are found in the DuckDB file.

        Examples
        --------
        >>> from nosible import ResultSet, Result
        >>> results = [
        ...     Result(url="https://example.com", title="Example Domain", similarity=0.95),
        ...     Result(url="https://openai.com", title="OpenAI", similarity=0.99),
        ... ]
        >>> search_results = ResultSet(results)
        >>> db_path = search_results.to_duckdb(file_path="results.duckdb", table_name="search_results")
        >>> loaded = ResultSet.from_duckdb("results.duckdb")
        >>> isinstance(loaded, ResultSet)
        True
        >>> len(loaded)
        2
        >>> loaded[0].title
        'Example Domain'
        """
        try:
            con = duckdb.connect(file_path, read_only=True)
        except Exception as e:
            raise RuntimeError(f"Failed to connect to DuckDB file '{file_path}': {e}") from e
        try:
            tables = con.execute("SHOW TABLES").fetchall()
            if not tables:
                raise ValueError(f"No tables found in DuckDB file '{file_path}'.")
            table_name = tables[0][0]
            try:
                arrow_table = con.execute(f"SELECT * FROM {table_name}").arrow()
            except Exception as e:
                raise RuntimeError(
                    f"Failed to fetch data from table '{table_name}' in DuckDB file '{file_path}': {e}"
                ) from e
            df = pl.from_arrow(arrow_table)
        except Exception as e:
            raise RuntimeError(f"Failed to read from DuckDB file '{file_path}': {e}") from e
        finally:
            con.close()
        try:
            return cls.from_polars(df)
        except Exception as e:
            raise RuntimeError(f"Failed to create ResultSet from DuckDB data in '{file_path}': {e}") from e

    @classmethod
    def from_dicts(cls, dicts: list[dict]) -> ResultSet:
        """
        Create a ResultSet instance from a list of dictionaries.

        Each dictionary in the input list should contain keys corresponding to the
        fields of a Result. This method will attempt to construct a Result
        object from each dictionary and collect them into a ResultSet container.

        Parameters
        ----------
        dicts : list of dict
            List where each element is a dictionary representing a Result.

        Returns
        -------
        ResultSet
            An instance containing all successfully parsed Result objects.

        Raises
        ------
        ValueError
            If any dictionary cannot be parsed into a Result.

        Examples
        --------
        >>> from nosible import ResultSet
        >>> dicts = [
        ...     {"url": "https://example.com", "title": "Example Domain", "url_hash": "abc123"},
        ...     {"url": "https://openai.com", "title": "OpenAI", "url_hash": "def456"},
        ... ]
        >>> results = ResultSet.from_dicts(dicts)
        >>> len(results)
        2
        >>> results[0].title
        'Example Domain'
        """
        results = []
        for d in dicts:
            try:
                # Try to get similarity first else go into semantics and get similarity
                result = Result(
                    url=d.get("url"),
                    title=d.get("title"),
                    description=d.get("description"),
                    netloc=d.get("netloc"),
                    published=d.get("published"),
                    visited=d.get("visited"),
                    author=d.get("author"),
                    content=d.get("content"),
                    language=d.get("language"),
                    similarity=d.get("similarity", d.get("semantics", {}).get("similarity")),
                    url_hash=d.get("url_hash"),
                )
                results.append(result)
            except Exception as e:
                raise ValueError(f"Error parsing dictionary into Result: {d}\n{e}") from e
        return cls(results)

    @classmethod
    def from_dict(cls, data: dict | list) -> ResultSet:
        """
        Create a ResultSet instance from a dictionary or a list of dictionaries.

        This method allows for flexible construction of ResultSet from either a single
        dictionary representing one search result, or a list of such dictionaries.

        Parameters
        ----------
        data : dict or list of dict
            A single dictionary or a list of dictionaries, where each dictionary contains
            keys corresponding to Result fields.

        Returns
        -------
        ResultSet
            An instance containing one or more Result objects parsed from the input.

        Raises
        ------
        ValueError
            If the input is neither a dictionary nor a list of dictionaries.
        RuntimeError
            If parsing the input fails.

        Examples
        --------
        >>> from nosible import ResultSet
        >>> # From a single dictionary
        >>> single = {"url": "https://example.com", "title": "Example", "url_hash": "abc123"}
        >>> results = ResultSet.from_dict(single)
        >>> isinstance(results, ResultSet)
        True
        >>> len(results)
        1
        >>> results[0].title
        'Example'
        >>> # From a list of dictionaries
        >>> data = [
        ...     {"url": "https://a.com", "title": "A", "url_hash": "a1"},
        ...     {"url": "https://b.com", "title": "B", "url_hash": "b2"},
        ... ]
        >>> results = ResultSet.from_dict(data)
        >>> len(results)
        2
        >>> results[1].title
        'B'
        """
        if isinstance(data, list):
            try:
                return cls.from_dicts(data)
            except Exception as e:
                raise RuntimeError(f"Failed to create ResultSet from list of dicts: {e}") from e
        elif isinstance(data, dict):
            try:
                return cls.from_dicts([data])
            except Exception as e:
                raise RuntimeError(f"Failed to create ResultSet from dict: {e}") from e
        else:
            raise ValueError("Input must be a list of dictionaries or a single dictionary.")

    def close(self) -> None:
        """
        Explicitly release any held resources.
        """
        # TODO: cleanup handles, sessions, etc.
        pass


if __name__ == "__main__":
    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE)
    print("All tests passed!")



from dataclasses import asdict, dataclass, field

from nosible.classes.snippet_set import SnippetSet
from nosible.utils.json_tools import json_dumps, json_loads


@dataclass(init=True, repr=True, eq=True, frozen=True)
class WebPageData:
    """
    A data container for all extracted and processed information about a web page.

    Parameters
    ----------
    full_text : str or None
        The full textual content of the web page, or None if not available.
    languages : dict
        A dictionary mapping language codes to their probabilities or counts, representing detected languages.
    metadata : dict
        Metadata extracted from the web page, such as description, keywords, author, etc.
    page : dict
        Page-specific details, such as title, canonical URL, and other page-level information.
    request : dict
        Information about the HTTP request and response, such as headers, status code, and timing.
    snippets : SnippetSet
        A set of extracted text snippets or highlights from the page, wrapped in a SnippetSet object.
    statistics : dict
        Statistical information about the page, such as word count, sentence count, or other computed metrics.
    structured : list
        A list of structured data objects (e.g., schema.org, OpenGraph) extracted from the page.
    url_tree : dict
        A hierarchical representation of the URL structure, such as breadcrumbs or navigation paths.

    Examples
    --------
    >>> data = WebPageData(languages={"en": 1}, metadata={"description": "Example"})
    >>> data.languages
    {'en': 1}
    >>> data.metadata
    {'description': 'Example'}
    """

    companies: list = None
    """A list of companies mentioned in the webpage, if applicable. (GKIDS)"""
    full_text: str = None
    """The full text content of the webpage."""
    languages: dict = None
    """Detected languages and their probabilities or counts."""
    metadata: dict = None
    """Metadata extracted from the webpage (e.g., description, keywords)."""
    page: dict = None
    """Page-specific details such as title, canonical URL, etc."""
    request: dict = None
    """Information about the HTTP request/response."""
    snippets: SnippetSet = field(init=True, default_factory=SnippetSet)
    """Extracted text snippets or highlights from the page."""
    statistics: dict = None
    """Statistical information about the page (e.g., word count)."""
    structured: list = None
    """Structured data (e.g., schema.org, OpenGraph)."""
    url_tree: dict = None
    """Hierarchical representation of the URL structure."""

    def __str__(self):
        """Return a string representation of the WebPageData.

        Returns
        -------
        str
            A string representation of the WebPageData instance, including languages, metadata, and other fields.
        """
        return (
            f"WebPageData(languages={self.languages}, metadata={self.metadata}, "
            f"page={self.page}, request={self.request}, snippets={self.snippets}, "
            f"statistics={self.statistics}, structured={self.structured}, url_tree={self.url_tree})"
        )

    def to_dict(self) -> dict:
        """
        Convert the WebPageData instance to a dictionary.

        Returns
        -------
        dict
            A dictionary containing all fields of the WebPageData.

        Examples
        --------
        >>> data = WebPageData(full_text="Example", languages={"en": 1}, metadata={"description": "Example"})
        >>> d = data.to_dict()
        >>> isinstance(d, dict)
        True
        >>> d["languages"] == {"en": 1}
        True
        """
        data = asdict(self)
        # snippets is still a SnippetSet instance, so convert it:
        data["snippets"] = self.snippets.to_dict()
        return data

    def to_json(self) -> str:
        """
        Convert the WebPageData to a JSON string representation.

        Returns
        -------
        str
            JSON string containing all fields of the WebPageData.

        Examples
        --------
        >>> data = WebPageData(languages={"en": 1}, metadata={"description": "Example"})
        >>> json_str = data.to_json()
        >>> isinstance(json_str, str)
        True
        """
        return json_dumps(self.to_dict())

    def save(self, path: str) -> None:
        """
        Save the WebPageData to a JSON file.

        Parameters
        ----------
        path : str
            Path to the file where the WebPageData will be saved.

        Examples
        --------
        >>> data = WebPageData(languages={"en": 1}, metadata={"description": "Example"})
        >>> data.save("test_webpage.json")
        >>> with open("test_webpage.json", "r", encoding="utf-8") as f:
        ...     content = f.read()
        >>> import json
        >>> d = json.loads(content)
        >>> d["languages"]
        {'en': 1}
        >>> d["metadata"]
        {'description': 'Example'}
        """
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.to_json())

    @classmethod
    def from_json(cls, data: str) -> "WebPageData":
        """
        Create a WebPageData instance from a JSON string.

        Parameters
        ----------
        data : str
            JSON string containing fields to initialize the WebPageData.

        Returns
        -------
        WebPageData
            An instance of WebPageData initialized with the provided JSON data.

        Examples
        --------
        >>> json_str = '{"languages": {"en": 1}, "metadata": {"description": "Example"}}'
        >>> webpage_data = WebPageData.from_json(json_str)
        >>> isinstance(webpage_data, WebPageData)
        True
        >>> webpage_data.languages
        {'en': 1}
        """
        data_dict = json_loads(data)
        # Handle snippets separately to avoid passing it twice
        snippets_data = data_dict.pop("snippets", None)
        if snippets_data is not None:
            data_dict["snippets"] = SnippetSet.from_dict(snippets_data)
        return cls(**data_dict)

    @classmethod
    def load(cls, path: str) -> "WebPageData":
        """
        Create a WebPageData instance from a JSON file.

        Parameters
        ----------
        path : str
            Path to the JSON file containing fields to initialize the WebPageData.

        Returns
        -------
        WebPageData
            An instance of WebPageData initialized with the provided data.

        Examples
        --------
        >>> data = WebPageData(languages={"en": 1}, metadata={"description": "Example"})
        >>> data.save("test_webpage.json")
        >>> loaded = WebPageData.load("test_webpage.json")
        >>> isinstance(loaded, WebPageData)
        True
        >>> loaded.languages
        {'en': 1}
        """
        with open(path, encoding="utf-8") as f:
            data = json_loads(f.read())
        # Handle snippets separately to avoid passing it twice
        snippets_data = data.pop("snippets", None)
        if snippets_data is not None:
            data["snippets"] = SnippetSet.from_dict(snippets_data)
        return cls(**data)



from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass, field

from nosible.classes.snippet import Snippet
from nosible.utils.json_tools import json_dumps


@dataclass()
class SnippetSet(Iterator[Snippet]):
    """
    An iterator and container for a collection of Snippet objects.
    This class allows iteration over, indexing into, and serialization of a set of Snippet objects.
    It supports initialization from a dictionary of snippet data, and provides
    methods for converting the collection to dictionary and JSON representations.

    Parameters
    ----------
    snippets : dict
        A dictionary where keys are snippet hashes and values are dictionaries of snippet attributes.

    Examples
    --------
    >>> snippets_data = {
    ...     "hash1": {"content": "Example snippet", "snippet_hash": "hash1"},
    ...     "hash2": {"content": "Another snippet", "snippet_hash": "hash2"},
    ... }
    >>> snippets = SnippetSet().from_dict(snippets_data)
    >>> for snippet in snippets:
    ...     print(snippet.content)
    Example snippet
    Another snippet
    """

    snippets: list[Snippet] = field(default_factory=list)
    """ List of `Snippet` objects contained in this ResultSet."""
    _index: int = field(default=0, init=False, repr=False, compare=False)
    """ Internal index for iteration over snippets."""

    def __iter__(self) -> SnippetSet:
        """
        Reset iteration and return self.

        Returns
        -------
        ResultSet
            Iterator over the ResultSet instance.
        """
        object.__setattr__(self, "_index", 0)
        return self

    def __next__(self) -> Snippet:
        """
        Returns the next Result in the sequence.

        Returns
        -------
        Result
            The next Result object in the sequence.
        Raises
        ------
        StopIteration
            If the end of the sequence is reached.
        """
        if self._index < len(self.snippets):
            item = self.snippets[self._index]
            object.__setattr__(self, "_index", self._index + 1)
            return item
        raise StopIteration

    def __len__(self) -> int:
        """
        Returns the number of snippets in the collection.

        Returns
        -------
        int
            The number of snippets.
        """
        return len(self.snippets)

    def __getitem__(self, index: int) -> Snippet:
        """
        Returns the Snippet at the specified index.

        Parameters
        ----------
        index : int
            The index of the snippet to retrieve.

        Returns
        -------
        Snippet
            The snippet at the specified index.

        Raises
        ------
        IndexError
            If the index is out of range.
        """
        if 0 <= index < len(self.snippets):
            return self.snippets[index]
        raise IndexError(f"Index {index} out of range for SnippetSet of length {len(self.snippets)}.")

    def __str__(self):
        """
        Print the content of all snippets in the set.
        Returns
        -------
        str
        """
        return "\n".join(str(s) for s in self)

    def to_dict(self) -> dict:
        """
        Convert the SnippetSet to a dictionary representation.

        Returns
        -------
        dict
            Dictionary containing all snippets indexed by their hash.

        Examples
        --------
        >>> snippets_data = {"hash1": {"content": "Example snippet", "snippet_hash": "hash1"}}
        >>> snippets = SnippetSet().from_dict(snippets_data)
        >>> snippets_dict = snippets.to_dict()
        >>> isinstance(snippets_dict, dict)
        True
        """
        return {s.snippet_hash: s.to_dict() for s in self.snippets} if self.snippets else {}

    def to_json(self) -> str:
        """
        Convert the SnippetSet to a JSON string representation.

        Returns
        -------
        str
            JSON string containing all snippets indexed by their hash.

        Examples
        --------
        >>> snippets_data = {"hash1": {"content": "Example snippet", "snippet_hash": "hash1"}}
        >>> snippets = SnippetSet().from_dict(snippets_data)
        >>> json_str = snippets.to_json()
        >>> isinstance(json_str, str)
        True
        """
        return json_dumps(self.to_dict())

    @classmethod
    def from_dict(cls, data: dict) -> SnippetSet:
        """
        Create a SnippetSet instance from a dictionary.

        Parameters
        ----------
        data : dict
            Dictionary containing snippet data.

        Returns
        -------
        SnippetSet
            An instance of SnippetSet populated with the provided data.

        Examples
        --------
        >>> snippets_data = {"hash1": {"content": "Example snippet", "snippet_hash": "hash1"}}
        >>> snippets = SnippetSet.from_dict(snippets_data)
        >>> isinstance(snippets, SnippetSet)
        True
        """
        return cls([Snippet.from_dict(s) for s in data.values()])



from typing import Union

try:
    import orjson

    _use_orjson = True
except ImportError:
    import json

    _use_orjson = False

# --------------------------------------------------------------------------------------------------------------
# Utility functions for JSON serialization/deserialization
# --------------------------------------------------------------------------------------------------------------


def json_dumps(obj: object) -> str:
    """
    Returns a JSON byte-string if using orjson, else a unicode str.

    Parameters
    ----------
    obj : object
        Object to serialize; must be JSON-serializable.

    Returns
    -------
    Union[bytes, str]
        If `orjson` is available (`_use_orjson` is True), returns a UTF-8 decoded
        unicode `str`; otherwise, returns a JSON `str` via the standard `json.dumps`.

    Raises
    ------
    RuntimeError
        If serialization fails for any reason.

    Examples
    --------
    # Standard dict serialization (no orjson)
    >>> _use_orjson = False
    >>> json_dumps({"a": 1})
    '{"a":1}'

    # List serialization
    >>> _use_orjson = False
    >>> json_dumps([1, 2, 3])
    '[1,2,3]'

    # orjson path returns unicode str
    >>> import orjson
    >>> _use_orjson = True
    >>> orjson.dumps = lambda o: b'{"b":2}'
    >>> json_dumps({"b": 2})
    '{"b":2}'

    # Non-str dict keys are coerced to str when using orjson
    >>> _use_orjson = True
    >>> orjson.dumps = lambda o: b'{"1":"one"}'
    >>> json_dumps({1: "one"})
    '{"1":"one"}'

    # Error path: un-serializable object
    >>> _use_orjson = False
    >>> class Bad: pass
    >>> try:
    ...     json_dumps(Bad())
    ... except RuntimeError as e:
    ...     "Failed to serialize object to JSON" in str(e)
    '{"1":"one"}'
    """
    try:
        if _use_orjson:
            # Ensure all dict keys are str for orjson
            def ensure_str_keys(o):
                if isinstance(o, dict):
                    return {str(k): ensure_str_keys(v) for k, v in o.items()}
                if isinstance(o, list):
                    return [ensure_str_keys(i) for i in o]
                return o

            obj = ensure_str_keys(obj)
            return orjson.dumps(obj).decode("utf-8")  # decode to str for consistency
        return json.dumps(obj)
    except Exception as e:
        # Optionally, you can log the error here
        raise RuntimeError(f"Failed to serialize object to JSON: {e}") from e


def json_loads(s: Union[bytes, str]) -> dict:
    """
    Accept both bytes (from orjson) and str (from json.loads).

    Parameters
    ----------
    s : Union[bytes, str]
        JSON data to deserialize. Can be `bytes` (e.g., from `orjson.dumps`)
        or a unicode `str`.

    Returns
    -------
    dict
        The deserialized Python dictionary.

    Raises
    ------
    RuntimeError
        If deserialization fails (invalid JSON or other error).

    Examples
    --------
    # Standard library path (disable orjson)
    >>> _use_orjson = False
    >>> json_dumps({'a': 1})
    '{"1":"one"}'

    # orjson path with monkey-patched dumping
    >>> _use_orjson = True
    >>> orjson.dumps = lambda o: b'{"b": 2}'
    >>> json_dumps({'b': 2})
    '{"b": 2}'

    # Convert non-str dict keys to str when using orjson
    >>> _use_orjson = True
    >>> orjson.dumps = lambda o: b'{"1": "one"}'
    >>> json_dumps({1: 'one'})
    '{"1": "one"}'

    # Error path: object not serializable
    >>> _use_orjson = False
    >>> class Bad: pass
    >>> try:
    ...     json_dumps(Bad())
    ... except RuntimeError as e:
    ...     "Failed to serialize" in str(e)
    '{"1": "one"}'
    """
    try:
        if _use_orjson:
            # orjson.loads accepts both bytes and str today
            return orjson.loads(s)
        # If we accidentally passed bytes, decode to str first
        if isinstance(s, (bytes, bytearray)):
            s = s.decode("utf-8")
        return json.loads(s)
    except Exception as e:
        # Optionally, you can log the error here
        raise RuntimeError(f"Failed to deserialize JSON: {e}") from e


def print_dict(dict: dict) -> str:
    """
    Print a dictionary in a readable format.

    Parameters
    ----------
    d : dict
        The dictionary to print.

    Returns
    -------
    None
        This function does not return anything; it prints the dictionary to stdout.
    """
    if _use_orjson:
        return orjson.dumps(dict, option=orjson.OPT_INDENT_2).decode("utf-8")
    return json.dumps(dict, indent=2)
